# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This modules contains low level image processing functions. These functions
are implemented in C for speed reasons. Usually one will not need to use these
directly, instead modules like "recognize" or "surface" use them to load and
analyze the image data.
"""

import os
import sys
import cairo

from sdaps import paths
from sdaps import defs
from sdaps import log

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

if paths.local_run:
    # image.so liegt in lib_build_dir/image/
    __path__.append(os.path.join(paths.lib_build_dir, 'image'))

# If SDAPS is installed, then the image.so file is in the current directory.
# Simply importing it without changes to the paths will work.

try:
    from .image import *
except ImportError as e:
    print(e)
    log.error(_("It appears you have not build the C extension. Please run \"./setup.py build\" in the toplevel directory."))
    sys.exit(1)

set_magic_values(defs.corner_mark_min_length,
                 defs.corner_mark_max_length,
                 defs.image_line_width,
                 defs.corner_mark_search_distance,
                 defs.image_line_coverage)

# Offset into corners, not valid for C API which has +1 as offset!
TOP_LEFT = 0
TOP_RIGHT = 1
BOTTOM_RIGHT = 2
BOTTOM_LEFT = 3

def calculate_matrix(surface, matrix, mm_x, mm_y, mm_width, mm_height):
    """Detect the transformation matrix

    This runs a detection for the corner marks that denote the bounding box
    given by mm_x, mm_y, mm_width, mm_height in the surface. The passed matrix
    is used for to estimate the resolution of the image.

    This function returns a new cairo matrix or raises an error otherwise.
    """
    found_corners = 0

    corners = []

    for i in (1, 2, 3, 4):
        try:
            corners.append(find_corner_marker (surface, matrix, i))
            found_corners += 1
        except:
            corners.append(None)

    # We need at least 3 corners to do anything
    assert found_corners >= 3

    return matrix_from_corners_2d(corners, mm_x, mm_y, mm_width, mm_height)


def matrix_from_corners_2d(corners, mm_x, mm_y, mm_width, mm_height):
    """Calculate the transformation matrix from bounding box corners

    Given the list of corners in the order top left, top right, bottom right,
    bottom left in px space and the bounding box that is defined by these
    corners in mm space, calcualte the corresponding px to mm transformation
    matrix.

    This function returns a new cairo matrix or raises an error otherwise.
    """

    assert len(corners) == 4
    assert corners.count(None) <= 1

    # Calculate a new "perfect" location for the undefined corner
    if corners[TOP_LEFT] is None:
        corners[TOP_LEFT] = (
                corners[BOTTOM_LEFT][0] - corners[BOTTOM_RIGHT][0] + corners[TOP_RIGHT][0],
                corners[TOP_RIGHT][1] - corners[BOTTOM_RIGHT][1] + corners[BOTTOM_LEFT][1]
            )
    if corners[TOP_RIGHT] is None:
        corners[TOP_RIGHT] = (
                corners[BOTTOM_RIGHT][0] - corners[BOTTOM_LEFT][0] + corners[TOP_LEFT][0],
                corners[TOP_LEFT][1] - corners[BOTTOM_LEFT][1] + corners[BOTTOM_RIGHT][1]
            )
    if corners[BOTTOM_LEFT] is None:
        corners[BOTTOM_LEFT] = (
                corners[TOP_LEFT][0] - corners[TOP_RIGHT][0] + corners[BOTTOM_RIGHT][0],
                corners[BOTTOM_RIGHT][1] - corners[TOP_RIGHT][1] + corners[TOP_LEFT][1]
            )
    if corners[BOTTOM_RIGHT] is None:
        corners[BOTTOM_RIGHT] = (
                corners[TOP_RIGHT][0] - corners[TOP_LEFT][0] + corners[BOTTOM_LEFT][0],
                corners[BOTTOM_LEFT][1] - corners[TOP_LEFT][1] + corners[TOP_RIGHT][1]
            )

    # X-Axis
    dx = ((corners[TOP_RIGHT][0] - corners[TOP_LEFT][0]) + (corners[BOTTOM_RIGHT][0] - corners[BOTTOM_LEFT][0])) / 2
    dy = ((corners[TOP_RIGHT][1] - corners[TOP_LEFT][1]) + (corners[BOTTOM_RIGHT][1] - corners[BOTTOM_LEFT][1])) / 2

    xx = dx / mm_width
    yx = dy / mm_width

    # y-Axis
    dx = ((corners[BOTTOM_RIGHT][0] - corners[TOP_RIGHT][0]) + (corners[BOTTOM_LEFT][0] - corners[TOP_LEFT][0])) / 2
    dy = ((corners[BOTTOM_RIGHT][1] - corners[TOP_RIGHT][1]) + (corners[BOTTOM_LEFT][1] - corners[TOP_LEFT][1])) / 2

    xy = dx / mm_height
    yy = dy / mm_height

    # Center everything between the markers
    x0 = (corners[BOTTOM_LEFT][0] + corners[BOTTOM_RIGHT][0] + corners[TOP_LEFT][0] + corners[TOP_RIGHT][0]) / 4
    y0 = (corners[BOTTOM_LEFT][1] + corners[BOTTOM_RIGHT][1] + corners[TOP_LEFT][1] + corners[TOP_RIGHT][1]) / 4

    x_center = mm_width / 2 + mm_x
    y_center = mm_height / 2 + mm_y

    dx = x_center * xx + y_center * xy
    dy = x_center * yx + y_center * yy

    x0 -= dx
    y0 -= dy

    m_new = cairo.Matrix(xx, yx, xy, yy, x0, y0)
    m_new.invert()

    return m_new



# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2010, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sdaps import model
from sdaps import script
from sdaps import defs

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext


parser = script.add_project_subparser("setup",
    help=_("Create a new survey using a LaTeX document."),
    description=_("""Create a new survey from a LaTeX document. You need to
    be using the SDAPS class. All the metadata and options for the project
    can be set inside the LaTeX document."""))

parser.add_argument('questionnaire.tex',
    help=_("The LaTeX Document"))
parser.add_argument('-a', '--add',
    help=_("Additional files that are required by the LaTeX document and need to be copied into the project directory."),
    action='append', default=[])
parser.add_argument('-e', '--engine',
    help=_("The engine to use to compile LaTeX documents."),
    default=defs.latex_engine)
parser.add_argument('additional_questions',
    nargs='?',
    help=_("Additional questions that are not part of the questionnaire."))

@script.connect(parser)
def setup(cmdline):
    from sdaps import setuptex

    return setuptex.setup(cmdline['project'],
                          cmdline['questionnaire.tex'],
                          cmdline['engine'],
                          cmdline['additional_questions'],
                          cmdline['add'])





#!/usr/bin/env python3
# -*- coding: utf8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright (C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright (C) 2012, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os

# Use the following and local_run=True below to run without installing SDAPS
#sys.path.append(os.path.join(os.path.dirname(sys.argv[0]), '..'))

import sdaps
#sdaps.init(local_run=True)
sdaps.init()

from sdaps import model
from sdaps import image

# Assume the first argument is a survey
survey = model.survey.Survey.load(sys.argv[1])

# We need the recognize buddies, as they are able to identify the data
from sdaps.recognize import buddies

# A sheet object to attach the images to
sheet = model.sheet.Sheet()
survey.add_sheet(sheet)

images = []

for file in sys.argv[2:]:
    num_pages = image.get_tiff_page_count(file)
    for page in range(num_pages):
        images.append((file, page))

if len(images) == 0:
    # No images, simply exit again.
    sys.exit(1)


def add_image(survey, tiff, page):
    img = model.sheet.Image()
    survey.sheet.add_image(img)
    # SDAPS assumes a relative path from the survey directory
    img.filename = os.path.relpath(os.path.abspath(tiff), survey.survey_dir)
    img.orig_name = tiff
    img.tiff_page = page

while images:
    # Simply drop the list of images again.
    sheet.images = []

    add_image(survey, *images.pop(0))

    if survey.defs.duplex:
        add_image(survey, *images.pop(0))
    else:
        add_image(survey, 'DUMMY', -1)

    # Run the recognition algorithm over the given images
    survey.questionnaire.recognize.recognize()

    for qobject in survey.questionnaire.qobjects:
        if isinstance(qobject, model.questionnaire.Question):
            # Only print data if an image for the page has been loaded
            if survey.sheet.get_page_image(qobject.page_number) is None:
                continue
            for box in qobject.boxes:
                print("%s,%s,%s,%s,%i,%f" % (
                    survey.sheet.global_id,
                    survey.sheet.survey_id,
                    survey.sheet.questionnaire_id,
                    '_'.join([str(num) for num in box.id]),
                    int(box.data.state),
                    float(box.data.quality)))
    print()

# And, we simply quit, ie. we don't save the survey





import sys
import os
import tempfile
import shutil

from sdaps import log
from sdaps import paths
from sdaps import defs

from sdaps.utils import latex
from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

def create_stamp_pdf(survey, output_filename, questionnaire_ids):

    if questionnaire_ids is None:
        log.warn(_("There should be no need to stamp a SDAPS Project that uses LaTeX and does not have different questionnaire IDs printed on each sheet.\nI am going to do so anyways."))

    # Temporary directory for TeX files.
    tmpdir = tempfile.mkdtemp(prefix='sdaps-stamp-')

    try:
        latex.write_override(survey, os.path.join(tmpdir, 'sdaps.opt'), questionnaire_ids=questionnaire_ids)

        print(_("Running %s now multiple times to generate the stamped questionnaire.") % survey.defs.engine)
        latex.compile(survey.defs.engine, 'questionnaire.tex', tmpdir, inputs=[os.path.abspath(survey.path())])

        if not os.path.exists(os.path.join(tmpdir, 'questionnaire.pdf')):
            log.error(_("Error running \"%s\" to compile the LaTeX file.") % defs.latex_engine)
            raise AssertionError('PDF file not generated')

        shutil.move(os.path.join(tmpdir, 'questionnaire.pdf'), output_filename)

    except:
        log.error(_("An error occured during creation of the report. Temporary files left in '%s'." % tmpdir))

        raise

    shutil.rmtree(tmpdir)



# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import tempfile

from sdaps import model
from sdaps import script
from sdaps import log

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext


parser = script.add_project_subparser("add",
    help=_("Add scanned questionnaires to the survey."),
    description=_("""This command is used to add scanned images to the survey.
    The image data needs to be a (multipage) 300dpi monochrome TIFF file. You
    may choose not to copy the data into the project directory. In that case
    the data will be referenced using a relative path."""))

parser.add_argument('--convert',
    help=_("Convert given files and add the result."),
    dest="convert",
    action="store_true",
    default=False)
parser.add_argument('--3d-transform',
    help=_("""Do a 3D-transformation after finding the corner marks. If the
        corner marks are not found then the image will be added as-is."""),
    dest="transform",
    action="store_true",
    default=False)
parser.add_argument('--force',
    help=_("Force adding the images even if the page count is wrong (only use if you know what you are doing)."),
    action="store_true",
    default=False)
parser.add_argument('--copy',
    help=_("Copy the files into the directory (default)."),
    dest="copy",
    action="store_true",
    default=True)
parser.add_argument('--no-copy',
    help=_("Do not copy the files into the directory."),
    dest="copy",
    action="store_false")
parser.add_argument('--duplex',
    help=_("Images contain a duplex scan of a simplex questionnaire (default: simplex scan)."),
    dest="duplex",
    action="store_true",
    default=False)

parser.add_argument('images',
    help=_("A number of TIFF image files."),
    nargs='+')

@script.connect(parser)
@script.logfile
def add(cmdline):
    import sys
    from sdaps.add import add_image, check_image
    from sdaps import image

    error = False
    survey = model.survey.Survey.load(cmdline['project'])

    filelist = []
    deletelist = []

    if not cmdline['convert']:
        for file in cmdline['images']:
            filelist.append(file)

            if not check_image(survey, file, cmdline['duplex'], cmdline['force'], message=True):
                error=True
        if error:
            return
    else:
        if not cmdline['copy']:
            log.error(_("The --no-copy option is not compatible with --convert!"))
            return 1

        try:
            from sdaps.convert import convert_images
        except:
            log.error("Need to convert the images to monochrome TIFF, however the conversion module cannot be imported. You are likely missing the OpenCV dependency.")
            return 1

        print(_("Converting input files into a single temporary file."))

        tmp = tempfile.mktemp(suffix='.tif', prefix='sdaps-convert-')
        deletelist.append(tmp)
        filelist.append(tmp)

        # Run conversion
        # TODO: Allow 3D transformation here!
        try:
            convert_images(cmdline['images'], tmp, survey.defs.paper_width, survey.defs.paper_height, cmdline['transform'])

            if not check_image(survey, tmp, cmdline['duplex'], cmdline['force']):
                log.error(_("The page count of the created temporary file does not work with this survey."))
                raise AssertionError()

        except Exception as e:
            log.error(str(e))
            log.error(_("Running the conversion failed."))
            error = True
            raise

    if not error:
        for file in filelist:
            print(_('Processing %s') % file)

            add_image(survey, file, cmdline['duplex'], cmdline['force'], cmdline['copy'])

            print(_('Done'))

    for file in deletelist:
        try:
            os.unlink(file)
        except OSError:
            pass

    if error:
        return 1
    else:
        survey.save()
        return 0





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import io
from PIL import Image

from reportlab import pdfgen
from reportlab import platypus
from reportlab.lib import styles
from reportlab.lib import units
from reportlab.lib import pagesizes
from reportlab.lib import enums
from reportlab.lib import colors
from xml.sax.saxutils import escape

from sdaps import template
from sdaps import model
from sdaps import image

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

from . import flowables


mm = units.mm

stylesheet = dict(template.stylesheet)

stylesheet['Right'] = styles.ParagraphStyle(
    'Right',
    parent=stylesheet['Normal'],
    alignment=enums.TA_RIGHT,
)

stylesheet['Right_Highlight'] = styles.ParagraphStyle(
    'Right_Highlight',
    parent=stylesheet['Right'],
    textColor=colors.Color(255, 0, 0)
)

stylesheet['Normal_Highlight'] = styles.ParagraphStyle(
    'Normal_Highlight',
    parent=stylesheet['Normal'],
    textColor=colors.Color(255, 0, 0)
)


class Choice(platypus.Flowable):
    '''One answer of a choice
    '''

    box_width = 200
    box_height = 9
    box_depth = 5
    box_margin = 1
    value_width = 45
    gap = 6

    def __init__(self, answer, value, significant=0):
        platypus.Flowable.__init__(self)
        if significant:
            stylesheet_name = 'Right_Highlight'
        else:
            stylesheet_name = 'Right'
        self.answer = platypus.Paragraph(escape(answer), stylesheet[stylesheet_name])
        self.value = platypus.Paragraph(
            '%.2f %%' % (value * 100), stylesheet[stylesheet_name]
        )
        self.black_box = flowables.Box(
            value * self.box_width, self.box_height, self.box_depth,
            self.box_margin
        )
        self.black_box.transparent = 0
        self.black_box.fill = 1
        self.black_box.fill_color = (0.6, 0.6, 0.6)
        self.white_box = flowables.Box(
            (1 - value) * self.box_width, self.box_height, self.box_depth,
            self.box_margin
        )

    def wrap(self, available_width, available_height):
        self.width = available_width
        self.white_box.wrap(available_width, available_height)
        available_width -= self.white_box.width - self.white_box.cx
        self.black_box.wrap(available_width, available_height)
        available_width -= self.black_box.width
        available_width -= self.gap
        self.value.wrap(self.value_width, available_height)
        available_width -= self.value.width
        self.answer.wrap(available_width, available_height)
        self.height = max(
            self.answer.height, self.value.height,
            self.black_box.height + self.box_margin,
            self.white_box.height + self.box_margin
        )
        return self.width, self.height

    def draw(self):
        if 0:
            assert isinstance(self.canv, pdfgen.canvas.Canvas)
        self.answer.drawOn(
            self.canv,
            0,
            self.height / 2.0 - self.answer.height / 2.0
        )
        self.value.drawOn(
            self.canv,
            self.answer.width,
            self.height / 2.0 - self.value.height / 2.0
        )
        self.black_box.drawOn(
            self.canv,
            self.answer.width + self.value.width + self.gap,
            self.height / 2.0 - self.black_box.height / 2.0
        )
        self.white_box.drawOn(
            self.canv,
            self.answer.width + self.value.width + self.gap + self.black_box.width - self.black_box.cx,
            self.height / 2.0 - self.white_box.height / 2.0
        )


class Range(platypus.Flowable):
    '''
    -----                self.top_margin
    values(percent)    self.values_height
    -----                self.values_gap
    values(bars)        self.bars_height
    -----
    skala with mean        self.skala_height
    -----
    marks(1 - range)        self.marks_height
    -----
    '''

    margin = 6
    top_margin = 0
    left_margin = 12

    def __init__(self, range_min, range_max, values, answers, mean, standard_deviation, count, significant=0):
        platypus.Flowable.__init__(self)

        self.range_min = range_min
        self.range_max = range_max

        self.values = values
        self.mean = mean
        self.standard_deviation = standard_deviation
        self.count = count

        self.box_width = 40
        self.box_height = 60
        self.box_depth = 6

        self.mean_width = 2
        self.mean_height = 6

        self.values_height = 10
        self.values_gap = self.box_depth
        self.bars_height = max(self.values.values()) * self.box_height
        self.skala_height = self.mean_height
        self.marks_height = 10

        if significant:
            stylesheet_name = 'Normal_Highlight'
        else:
            stylesheet_name = 'Normal'

        self.answers_paragraph = \
            platypus.Paragraph(escape(' - '.join(answers)), stylesheet[stylesheet_name])
        self.count_paragraph = \
            platypus.Paragraph(_('Answers: %i') % self.count, stylesheet['Normal'])
        self.mean_paragraph = \
            platypus.Paragraph(_('Mean: %.2f') % self.mean, stylesheet['Normal'])
        self.stdd_paragraph = \
            platypus.Paragraph(_('Standard Deviation: %.2f') % self.standard_deviation, stylesheet['Normal'])

    def wrap(self, available_width, available_height):
        self.answers_paragraph.wrap(available_width, available_height)
        self.count_paragraph.wrap(available_width, available_height)
        self.mean_paragraph.wrap(available_width, available_height)
        self.stdd_paragraph.wrap(available_width, available_height)
        self.width = available_width # self.box_width * 5
        self.offset = self.width - self.box_width * len(self.values) - self.margin
        self.height = self.top_margin + self.values_height + self.values_gap + \
                      self.bars_height + self.skala_height + self.marks_height
        return self.width, self.height

    def draw(self):
        if 0:
            assert isinstance(self.canv, pdfgen.canvas.Canvas)
        self.canv.setFont("Times-Roman", 10)
        # mean
        mean = flowables.Box(self.mean_width, self.mean_height, self.box_depth)
        mean.transparent = 0
        mean.fill = 1
        mean.fill_color = (0.6, 0.6, 0.6)
        mean.drawOn(self.canv,
                    self.offset + (self.mean - 0.5) * self.box_width - self.mean_width / 2.0,
                    self.marks_height)
        # values

        for i, key in enumerate(range(self.range_min, self.range_max + 1)):
            if key in self.values:
                self.canv.drawCentredString(
                    self.offset + (i + 0.5) * self.box_width + self.box_depth * 0.5,
                    self.marks_height + self.skala_height + self.bars_height + self.values_gap,
                    '%.2f %%' % (self.values[key] * 100)
                )
        # bars
        for i, key in enumerate(range(self.range_min, self.range_max + 1)):
            if key in self.values:
                value = self.values[key]
            else:
                value = 0

            box = flowables.Box(self.box_width, value * self.box_height, self.box_depth)
            box.transparent = 0
            box.fill = 1
            box.fill_color = (0.6, 0.6, 0.6)
            box.drawOn(self.canv,
                       self.offset + i * self.box_width, self.marks_height + self.skala_height)
        # skala
        for i in range((self.range_max - self.range_min) * 10 + 1):
            if i % 10 == 0:
                self.canv.setLineWidth(0.2)
                self.canv.line(
                    self.offset + (i / 10.0 + 0.5) * self.box_width,
                    1 + self.marks_height,
                    self.offset + (i / 10.0 + 0.5) * self.box_width,
                    5 + self.marks_height
                )
            elif i % 5 == 0:
                self.canv.line(
                    self.offset + (i / 10.0 + 0.5) * self.box_width,
                    1.5 + self.marks_height,
                    self.offset + (i / 10.0 + 0.5) * self.box_width,
                    4.5 + self.marks_height
                )
            else:
                self.canv.line(
                    self.offset + (i / 10.0 + 0.5) * self.box_width,
                    2 + self.marks_height,
                    self.offset + (i / 10.0 + 0.5) * self.box_width,
                    4 + self.marks_height
                )
            if i % 10 == 0:
                self.canv.setLineWidth(0.1)
        # marks
        for i in range(1, len(self.values)+1):
            self.canv.drawCentredString(
                self.offset + (i - 0.5) * self.box_width, 0,
                '%i' % i
            )

        # statistics
        y_pos = self.marks_height + self.skala_height + self.bars_height + \
                self.values_gap + self.values_height
        self.answers_paragraph.drawOn(self.canv, self.left_margin, y_pos - 15)
        self.count_paragraph.drawOn(self.canv, self.left_margin, y_pos - 27)
        self.mean_paragraph.drawOn(self.canv, self.left_margin, y_pos - 39)
        self.stdd_paragraph.drawOn(self.canv, self.left_margin, y_pos - 51)


class Freeform(platypus.Flowable):

    cache = dict()

    def __init__(self, box):
        platypus.Flowable.__init__(self)

        assert isinstance(box, model.questionnaire.Textbox)
        assert box.data.state

        image = box.sheet.get_page_image(box.question.page_number)

        self.filename = box.question.questionnaire.survey.path(image.filename)
        self.tiff_page = image.tiff_page
        self.rotated = image.rotated

        mm_to_px = image.matrix.mm_to_px()
        x0, y0 = mm_to_px.transform_point(box.data.x, box.data.y)
        x1, y1 = mm_to_px.transform_point(box.data.x + box.data.width, box.data.y + box.data.height)

        self.bbox = (int(x0), int(y0), int(x1), int(y1))

        self.width = box.data.width * mm
        self.height = box.data.height * mm

    def wrap(self, available_width, available_height):
        self.available_width = available_width
        return self.available_width, self.height

    def draw(self):
        if 0:
            assert isinstance(self.canv, pdfgen.canvas.Canvas)
        if(self.filename, self.tiff_page, self.bbox) in self.cache:
            img = self.cache[(self.filename, self.tiff_page, self.bbox)]
        else:
            img = io.BytesIO(image.get_pbm(
                image.get_a1_from_tiff(
                    self.filename,
                    self.tiff_page,
                    self.rotated if self.rotated else False
                )
            ))
            img = Image.open(img).crop(self.bbox)
            self.cache[(self.filename, self.bbox)] = img
        self.canv.drawInlineImage(img, 0, 0, self.width, self.height)
        self.canv.setStrokeColorRGB(0.6, 0.6, 0.6)
        self.canv.line(0, 0, self.available_width, 0)
        self.canv.line(0, self.height, self.available_width, self.height)


class RawText(platypus.Paragraph):

    def __init__(self, text, *args, **kwargs):

        # Replace things like 

        text = escape(text)
        text = text.replace('\n', '<br/>')

        text = text

        platypus.Paragraph.__init__(self, text, *args, bulletText='•', **kwargs)




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import csv

from sdaps import clifilter
from sdaps import model

from . import buddies


def csvdata_export(survey, csvfile, image_writer, filter=None, export_images=False, export_question_images=False, export_quality=False, csvoptions={}):
    # compile clifilter
    filter = clifilter.clifilter(survey, filter)

    # export
    survey.questionnaire.csvdata.open_csv(csvfile, image_writer, export_images, export_question_images, export_quality, csvoptions)

    survey.iterate(
        survey.questionnaire.csvdata.export_data,
        filter,
    )

    survey.questionnaire.csvdata.export_finish()


def csvdata_import(survey, csvfile):
    csvreader = csv.DictReader(csvfile)

    for data in csvreader:
        survey.questionnaire.csvdata.import_data(data)

    csvfile.close()

    survey.save()





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Template
========

This modules contains templates for reportlab. These are used to create the
reports.
"""

from reportlab import platypus
from reportlab.lib import styles
from reportlab.lib import units
from reportlab.lib import pagesizes

mm = units.mm
PADDING = 15 * mm

class DocTemplate(platypus.BaseDocTemplate):

    def __init__(self, filename, title, metainfo={}, papersize=pagesizes.A4):
        platypus.BaseDocTemplate.__init__(
            self, filename,
            pagesize=papersize,
            leftMargin=0,
            rightMargin=0,
            topMargin=0,
            bottomMargin=0,
            **metainfo
        )
            #pageTemplates=[],
            #showBoundary=0,
            #allowSplitting=1,
            #title=None,
            #author=None,
            #_pageBreakQuick=1
        self.addPageTemplates(TitlePageTemplate(papersize, title))
        self.addPageTemplates(PageTemplate(papersize))


class TitlePageTemplate(platypus.PageTemplate):

    def __init__(self, papersize, title):
        self.title = title
        frames = [
            platypus.Frame(
                PADDING, papersize[1] / 2,
                papersize[0] - 2*PADDING, papersize[1] / 6,
                showBoundary=0
            ),
            platypus.Frame(
                PADDING, PADDING,
                papersize[0] - 2*PADDING, papersize[1] / 2 - PADDING,
                showBoundary=0
            ),
        ]
        platypus.PageTemplate.__init__(
            self,
            id='Title',
            frames=frames,
        )

    def beforeDrawPage(self, canvas, document):
        canvas.saveState()
        canvas.setFont('Times-Bold', 24)
        canvas.drawCentredString(
            document.width / 2.0,
            document.height - 50 * mm,
            self.title
        )
        canvas.restoreState()

    def afterDrawPage(self, canvas, document):
        pass


class PageTemplate(platypus.PageTemplate):

    def __init__(self, papersize):
        frames = [
            platypus.Frame(
                PADDING, PADDING,
                papersize[0] - 2*PADDING, papersize[1] - 2*PADDING,
                showBoundary=0
            )
        ]
        platypus.PageTemplate.__init__(
            self,
            id='Normal',
            frames=frames,
        )
        self.frames

    def beforeDrawPage(self, canvas, document):
        pass

    def afterDrawPage(self, canvas, document):
        pass


stylesheet = dict()

stylesheet['Normal'] = styles.ParagraphStyle(
    'Normal',
    fontName='Times-Roman',
    fontSize=10,
    leading=14,
)

stylesheet['Title'] = styles.ParagraphStyle(
    'Title',
    parent=stylesheet['Normal'],
    fontSize=18,
    leading=22,
    alignment=styles.TA_CENTER,
)


def story_title(survey, info=dict()):
    story = [
        platypus.Paragraph(str(line), stylesheet['Title'])
        for line in survey.title.split('\n')
    ]
    story += [
        platypus.FrameBreak(),
    ]

    keys = list(survey.info.keys())
    if keys:
        keys.sort()
        table = [
            [
                platypus.Paragraph(str(key), stylesheet['Normal']),
                platypus.Paragraph(str(survey.info[key]), stylesheet['Normal'])
            ]
            for key in keys
        ]
        story += [
            platypus.Table(table, colWidths=(50 * mm, None)),
        ]
    if info:
        story += [
            platypus.Spacer(0, 10 * mm)
        ]
        keys = list(info.keys())
        keys.sort()
        table = [
            [
                platypus.Paragraph(str(key), stylesheet['Normal']),
                platypus.Paragraph(str(info[key]), stylesheet['Normal'])
            ]
            for key in keys
        ]
        story += [
            platypus.Table(table, colWidths=(50 * mm, None)),
        ]

    story += [
        platypus.NextPageTemplate('Normal'),
        platypus.PageBreak()
    ]
    return story




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2012, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sdaps import model
from sdaps import script

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext


parser = script.add_project_subparser("reset",
    help=_("Reset project into original state"),
    description=_("""This command does a full reset of the project. All data
    will be discarded and only the empty project is left.
    """))


@script.connect(parser)
@script.logfile
def reset(cmdline):
    survey = model.survey.Survey.load(cmdline['project'])

    from sdaps import reset

    return reset.reset(survey)




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math

from reportlab import pdfgen
from reportlab import platypus
from reportlab.lib import styles
from reportlab.lib import units
from reportlab.lib import pagesizes
from reportlab.lib import enums

from sdaps import template


mm = units.mm

stylesheet = template.stylesheet

stylesheet['Right'] = styles.ParagraphStyle(
    'Right',
    parent=stylesheet['Normal'],
    alignment=enums.TA_RIGHT,
)


class Box(platypus.Flowable):
    '''3d box
    '''

    def __init__(self, a, b, c, margin=0):
        platypus.Flowable.__init__(self)
        self.a = float(a)
        self.b = float(b)
        self.c = float(c)
        self.margin = float(margin)
        self.alpha = 1.0 / 3.0 * math.pi
        self.cx = math.sin(self.alpha) * self.c
        self.cy = math.cos(self.alpha) * self.c
        self.fill = 0
        self.transparent = 1
        self.fill_color = (255, 255, 255)

    def wrap(self, available_width, available_height):
        self.width = self.a + self.cx
        self.height = self.b + self.cy + 2 * self.margin
        return self.width, self.height

    def draw(self):
        if 0:
            assert isinstance(self.canv, pdfgen.canvas.Canvas)
        self.canv.setLineJoin(1)
        self.canv.setLineWidth(0.2)
        self.canv.setFillColorRGB(*self.fill_color)
        # back
        if self.transparent:
            self.draw_front(self.cx, self.margin + self.cy)
        # left side
        if self.transparent:
            self.draw_side(0, self.margin)
        # bottom
        if self.transparent:
            self.draw_top(0, self.margin)
        # right side
        self.draw_side(self.a, self.margin)
        # top
        self.draw_top(0, self.margin + self.b)
        # front
        self.draw_front(0, self.margin)

    def draw_side(self, x, y):
        path = self.canv.beginPath()
        path.moveTo(x, y)
        path.lineTo(x + self.cx, y + self.cy)
        path.lineTo(x + self.cx, y + self.cy + self.b)
        path.lineTo(x, y + self.b)
        path.lineTo(x, y)
        self.canv.drawPath(path, fill=self.fill)

    def draw_top(self, x, y):
        path = self.canv.beginPath()
        path.moveTo(x, y)
        path.lineTo(x + self.a, y)
        path.lineTo(x + self.a + self.cx, y + self.cy)
        path.lineTo(x + self.cx, y + self.cy)
        path.lineTo(x, y)
        self.canv.drawPath(path, fill=self.fill)

    def draw_front(self, x, y):
        path = self.canv.beginPath()
        path.moveTo(x, y)
        path.lineTo(x + self.a, y)
        path.lineTo(x + self.a, y + self.b)
        path.lineTo(x, y + self.b)
        path.lineTo(x, y)
        self.canv.drawPath(path, fill=self.fill)


#class ChoiceBar(platypus.Flowable):

    #def __init__(self, value):
        #platypus.Flowable.__init__(self)
        #self.a = 200
        #self.b = 12
        #self.depth_x = 6
        #self.depth_y = 3
        #self.value = value

    #def wrap(self, available_width, available_height):
        #self.width = self.a + self.depth_x
        #self.height = self.height + self.depth_y
        #return self.width, self.height

    #def draw(self):
        #if 0: assert isinstance(self.canv, pdfgen.canvas.Canvas)
        #self.canv.setLineJoin(1)
        #self.canv.setLineWidth(0.1)
        #self.canv.setFillGray(0.5)
        ## filling
        #self.draw_side(self.value, 1)
        #self.draw_top(self.value, 1)
        #self.draw_front(self.value, 1)
        ## line
        #self.canv.line(self.a * self.value,                0,            self.a,                0)
        #self.canv.line(self.a * self.value + self.depth_x, self.depth_y, self.a + self.depth_x, self.depth_y)
        #self.canv.line(self.a * self.value,                self.b,                self.a,                self.b)
        #self.canv.line(self.a * self.value + self.depth_x, self.b + self.depth_y, self.a + self.depth_x, self.b + self.depth_y)
        ## bar
        #self.draw_side(1, 0)

    #def draw_front(self, value, fill):
        #path = self.canv.beginPath()
        #path.moveTo(0, 0)
        #path.lineTo(self.a * value, 0)
        #path.lineTo(self.a * value, self.b)
        #path.lineTo(0, self.b)
        #path.lineTo(0, 0)
        #self.canv.drawPath(path, fill=fill)

    #def draw_top(self, value, fill):
        #path = self.canv.beginPath()
        #path.moveTo(0, self.b)
        #path.lineTo(self.a * value, self.b)
        #path.lineTo(self.a * value + self.depth_x, self.b + self.depth_y)
        #path.lineTo(self.depth_x, self.b + self.depth_y)
        #path.lineTo(0, self.b)
        #self.canv.drawPath(path, fill=fill)

    #def draw_side(self, value, fill):
        #path = self.canv.beginPath()
        #path.moveTo(self.a * value, 0)
        #path.lineTo(self.a * value, self.b)
        #path.lineTo(self.a * value + self.depth_x, self.b + self.depth_y)
        #path.lineTo(self.a * value + self.depth_x, self.depth_y)
        #path.lineTo(self.a * value, 0)
        #self.canv.drawPath(path, fill=fill)


#class ChoiceAnswer(platypus.Flowable):

    #def __init__(self, answer, value):
        #platypus.Flowable.__init__(self)
        #self.answer = platypus.Paragraph(answer, template.stylesheet['Right'])
        #self.value = platypus.Paragraph(u'%.2f %%' %(value * 100), template.stylesheet['Right'])
        #self.bar = ChoiceBar(value)
        #self.gap = 3
        #self.value_width = 42

    #def wrap(self, available_width, available_height):
        #self.width = available_width
        #self.bar.wrap(self.width, available_height)
        #available_width -= self.bar.width
        #available_width -= self.gap
        #self.value.wrap(self.value_width, available_height)
        #available_width -= self.value.width
        #self.answer.wrap(available_width, available_height)
        #self.height = max(self.answer.height, self.bar.height, self.value.height)
        #return self.width, self.height


    #def draw(self):
        #if 0: assert isinstance(self.canv, pdfgen.canvas.Canvas)
        #self.answer.drawOn(self.canv, 0, 0)
        #self.value.drawOn(self.canv, self.answer.width, 0)
        #self.bar.drawOn(self.canv, self.answer.width + self.value.width + self.gap, 0)

#class MarkAnswer(platypus.Flowable):

    #def __init__(self):
        #platypus.Flowable.__init__(self)
        #self.box_width = 40
        #self.box_height = 60
        #self.box_depth = 6
        #self.mean_width = 2
        #self.mean_height = 6
        #self.margin = 6
        ##self.values = [0.2, 0.20, 0.2, 0.20, 0.20]
        #self.values = [0.25, 0.10, 0.35, 0.20, 0.10]
        #self.mean = sum([(mark + 1) * value for mark, value in enumerate(self.values)])


    #def wrap(self, available_width, available_height):
        #self.width = available_width #self.box_width * 5
        #self.offset = self.width - self.box_width * 5 - self.margin
        #self.height = max(self.values) * self.box_height + self.mean_height
        #return self.width, self.height

    #def draw(self):
        #if 0: assert isinstance(self.canv, pdfgen.canvas.Canvas)
        #self.canv.setLineJoin(1)
        #self.canv.setLineWidth(0.1)
        #self.canv.setFillGray(0.5)
        ## mean
        #Box(self.mean_width, self.mean_height, self.box_depth).drawOn(self.canv, self.offset +(self.mean - 0.5) * self.box_width - self.mean_width / 2.0, 0)
        ## boxes
        #for i, value in enumerate(self.values):
            #Box(self.box_width, value * self.box_height, self.box_depth).drawOn(self.canv, self.offset + i * self.box_width, self.mean_height)
        ## skala
        #for i in range(41):
            #if i % 10 == 0:
                #self.canv.setLineWidth(0.2)
                #self.canv.line(self.offset +(i / 10.0 + 0.5) * self.box_width, 1, self.offset +(i / 10.0 + 0.5) * self.box_width, 5)
            #elif i % 5 == 0:
                #self.canv.line(self.offset +(i / 10.0 + 0.5) * self.box_width, 1.5, self.offset +(i / 10.0 + 0.5) * self.box_width, 4.5)
            #else:
                #self.canv.line(self.offset +(i / 10.0 + 0.5) * self.box_width, 2, self.offset +(i / 10.0 + 0.5) * self.box_width, 4)
            #if i % 10 == 0:
                #self.canv.setLineWidth(0.1)





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
    sdaps - scripts for data acquisition with paper based surveys

    model/buddy - Registration center for buddies


Defining a buddy
================

    from sdaps import model

    class QObject(model.buddy.Buddy):

        __metaclass__ = model.buddy.Register
        obj_class = object class
        name = 'my_buddy'

    The buddy will be available as object.my_buddy

    Inside buddy class, the object is available as self.obj

'''


class Object(object):
    '''class which can have buddies'''

    def get_buddy(self, name):
        try:
            return getattr(self, '_%s_object_' % name)
        except AttributeError:
            setattr(self, '_%s_object_' % name, getattr(self, '_%s_class_' % name)(self))
            return getattr(self, '_%s_object_' % name)


class Register(type):
    '''metaclass to register the class as a buddy'''

    def __init__(cls, name, bases, dict):
        type.__init__(cls, name, bases, dict)
        assert issubclass(cls.obj_class, Object)
        setattr(
            cls.obj_class,
            '_%s_class_' % cls.name,
            cls
        )
        setattr(
            cls.obj_class,
            cls.name,
            property(lambda self: self.get_buddy(cls.name))
        )


class Buddy(object, metaclass=Register):
    '''base class for buddies'''
    obj_class = Object
    name = 'my_buddy'

    def __init__(self, obj):
        if 0:
            assert isinstance(obj, self.obj_class)
        self.obj = obj




#!/usr/bin/env python3
# -*- coding: utf8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright (C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright (C) 2012, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os

# Use the following and local_run=True below to run without installing SDAPS
#sys.path.append(os.path.join(os.path.dirname(sys.argv[0]), '..'))

import sdaps
#sdaps.init(local_run=True)
sdaps.init()

from sdaps import model
from sdaps import image

# We need a survey that has the correct definitions (paper size, duplex mode)
# Assume the first argument is a survey
survey = model.survey.Survey.load(sys.argv[1])

# We need the recognize buddies, as they are able to identify the data
from sdaps.recognize import buddies

# A sheet object to attach the images to
sheet = model.sheet.Sheet()
survey.add_sheet(sheet)

images = []

for file in sys.argv[2:]:
    num_pages = image.get_tiff_page_count(file)
    for page in range(num_pages):
        images.append((file, page))

if len(images) == 0:
    # No images, simply exit again.
    sys.exit(1)


def add_image(survey, tiff, page):
    img = model.sheet.Image()
    survey.sheet.add_image(img)
    # SDAPS assumes a relative path from the survey directory
    img.filename = os.path.relpath(os.path.abspath(tiff), survey.survey_dir)
    img.orig_name = tiff
    img.tiff_page = page

while images:
    # Simply drop the list of images again.
    sheet.images = []

    add_image(survey, *images.pop(0))

    if survey.defs.duplex:
        add_image(survey, *images.pop(0))

    sheet.recognize.recognize()

    for img in sheet.images:
        print(img.orig_name, img.tiff_page)
        print('\tPage:', img.page_number)
        print('\tRotated:', img.rotated)
        print('\tMatrix (px to mm):', img.raw_matrix)
        print('\tSurvey-ID:', sheet.survey_id)
        print('\tGlobal-ID:', sheet.global_id)
        print('\tQuestionnaire-ID:', sheet.questionnaire_id)
        print()

# And, we simply quit, ie. we don't save the survey




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess


def mimetype(filename):
    '''return the mimetype of the file as string or an error string,
    if the file does not exist or is not accesible.
    '''
    file = subprocess.Popen(
        ['file', '--brief', '--mime-type', filename],
        stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    stdout, stderr = file.communicate()
    if stdout:
        stdout = stdout.strip()
    return stdout.decode('ascii')




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from . import buddy
from . import data as datamodule
from . import db

class Sheet(buddy.Object):

    _save_attrs = {'data', 'images', 'survey_id',
                   'questionnaire_id', 'global_id', 'valid',
                   'quality', 'recognized', 'review_comment' }

    def __init__(self):
        self.survey = None
        self.data = dict()
        self.images = list()
        self.survey_id = None
        self.questionnaire_id = None
        self.global_id = None
        self.valid = 1
        self.quality = 1
        self.review_comment = None

        self.recognized = False

    def add_image(self, image):
        self.images.append(image)
        image.sheet = self

    def get_page_image(self, page):
        # Simply return the image for the requested page.
        # Note: We return the first one we find; this means in the error case
        #       that a page exists twice, we return the first one.
        for image in self.images:
            if image.page_number == page and image.survey_id == self.survey.survey_id:
                return image
        return None

    def reinit_state(self):
        for k, v in self.data.items():
            obj = self.survey.questionnaire.find_object(k)

            v._parent = obj

    @classmethod
    def __to_json_state__(cls, state):
        # json is unable to serialize our tuples as keys, so make strings with
        # '^' as a separator.
        orig_data = state['data']
        state['data'] = dict()

        for k, v in orig_data.items():
            state['data']['^'.join(str(_) for _ in k)] = v

    def __setstate__(self, data):
        # Attributes that may not (yet) be present in the database
        self.survey = None
        self.review_comment = None

        _tmp = data['data']
        data['data'] = dict()
        for k, v in _tmp.items():
            data['data'][tuple(int(_) for _ in k.split('^'))] = db.fromJson(v, datamodule)

        for k in range(len(data['images'])):
            data['images'][k] = db.fromJson(data['images'][k], Image)
            data['images'][k].sheet = self

        # Migrate old "verified" flag
        if 'verified' in data:
            for img in data['images']:
                img.verified = data['verified']
            del data['verified']

        self.__dict__.update(data)

    @property
    def verified(self):
        for img in self.images:
            if img.ignored:
                continue

            if not img.verified:
                return False

        return True

    @property
    def empty(self):
        for k, v in self.data.items():
            if not v.empty:
                return False

        return True

    @property
    def complete(self):
        """A boolean whether this sheet is complete in the sense that
        every page of the questionnaire has been identified.
        ie. it is false if there are missing pages"""

        # Simply retrieve every page, and see if it is not None
        for page in range(self.survey.questionnaire.page_count):
            if self.get_page_image(page + 1) is None:
                return False
        return True

    @property
    def dirty(self):
        if hasattr(self, '_dirty'):
             return True
        for d in self.data.values():
            if hasattr(d, '_dirty'):
                return True
        for i in self.images:
            if hasattr(i, '_dirty'):
                return True

    def _clear_dirty(self):
        if hasattr(self, '_dirty'):
             del self._dirty
        for d in self.data.values():
            if hasattr(d, '_dirty'):
                del d._dirty
        for i in self.images:
            if hasattr(i, '_dirty'):
                del i._dirty

    def __setattr__(self, attr, value):
        if attr.startswith('_') or attr == 'survey':
            object.__setattr__(self, attr, value)
            return

        # Nonexisting attributes should never be set.
        assert attr in self._save_attrs

        self.__setattr__('_dirty', True)

        # We need to fall back to "None" for __init__ to work.
        try:
            old_value = getattr(self, attr)
            force = False
        except AttributeError:
            old_value = None
            force = True

        if force or value != old_value:
            object.__setattr__(self, attr, value)
            # survey may be None if the sheet does not belong to a survey yet.
            if self.survey is not None:
                self.survey.questionnaire.notify_data_changed(None, None, attr, old_value)

class Image(buddy.Object):

    # Could make "sheet" a private property instead!
    _save_skip = {'sheet'}

    def __init__(self):
        self.sheet = None
        self.filename = str()
        self.tiff_page = 0
        self.rotated = 0
        self.raw_matrix = None
        self.page_number = None
        self.survey_id = None
        self.global_id = None
        self.questionnaire_id = None
        #: Whether the page should be ignored (because it is a blank back side)
        self.ignored = False
        self.verified = False

    def __setattr__(self, attr, value):
        if attr.startswith('_'):
            return object.__setattr__(self, attr, value)
        if attr != 'sheet':
            object.__setattr__(self, '_dirty', True)

        try:
            old_value = getattr(self, attr)
        except AttributeError:
            old_value = None

        object.__setattr__(self, attr, value)

        # survey may be None if the sheet does not belong to a survey yet.
        if hasattr(self, "sheet") and self.sheet is not None and self.sheet.survey is not None:
            self.sheet.survey.questionnaire.notify_data_changed(None, None, attr, old_value)




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from sdaps import model
from sdaps import script

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext


parser = script.add_project_subparser("gui",
    help=_("Launch a GUI. You can view and alter the (recognized) answers with it."),
    description=_("""This command launches a graphical user interface that can
    be used to correct answers. You need to run "recognize" before using it.
    """))

parser.add_argument('-f', '--filter',
    help=_("Filter to only show a partial dataset."))

@script.connect(parser)
@script.logfile
def gui(cmdline):
    from sdaps import gui

    survey = model.survey.Survey.load(cmdline['project'])

    return gui.gui(survey, cmdline)





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import csv

from sdaps import model

import os
import os.path
from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

class Questionnaire(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'csvdata'
    obj_class = model.questionnaire.Questionnaire

    def open_csv(self, csvfile, image_writer=None, export_images=False, export_question_images=False, export_quality=False, csv_options={}):
        header = ['questionnaire_id', 'global_id', 'empty', 'valid', 'recognized', 'review', 'verified']

        self.export_quality = export_quality

        self.image_writer = image_writer
        if image_writer is not None and (export_images or export_question_images):
            from sdaps.utils.image import ImageWriter

            self.export_images = export_images
            self.export_question_images = export_question_images

        else:
            self.export_images = None
            self.export_question_images = None

        for qobject in self.obj.qobjects:
            header.extend(qobject.csvdata.export_header())
        self.file = csvfile
        self.csv = csv.DictWriter(self.file, header, **csv_options)
        self.csv.writerow({value: value for value in header})

    def export_data(self):
        data = {
            'questionnaire_id': str(self.obj.sheet.questionnaire_id),
            'global_id': str(self.obj.sheet.global_id),
            'empty' : str(int(self.obj.sheet.empty)),
            'valid' : str(int(self.obj.sheet.valid)),
            'recognized' : str(int(self.obj.sheet.recognized)),
            'review' : str(self.obj.sheet.review_comment if self.obj.sheet.review_comment else ''),
            'verified' : str(int(self.obj.sheet.verified)),
        }
        for qobject in self.obj.qobjects:
            data.update(qobject.csvdata.export_data())
        self.csv.writerow(data)

    def export_finish(self):
        self.file.flush()
        del self.file
        del self.csv

    def import_data(self, data):
        try:
            self.obj.survey.goto_questionnaire_id(data['questionnaire_id'])
        except ValueError:
            print((_('Not importing unknown questionnaire ID "%s"') % data['questionnaire_id']))
            # The sheet does not exist
            # Ignore it
            pass
        else:
            if 'valid' in data:
                self.obj.sheet.valid = bool(int(data['valid']))
            if 'review' in data:
                self.obj.sheet.review_comment = data['review']
            if 'verified' in data:
                for img in self.obj.sheet.images:
                    img.verified = bool(int(data['verified']))

            for qobject in self.obj.qobjects:
                qobject.csvdata.import_data(data)


class QObject(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'csvdata'
    obj_class = model.questionnaire.QObject

    def export_header(self):
        header = [ self.obj.id_csv() + '_review' ]
        if self.obj.questionnaire.csvdata.export_question_images and self.obj.boxes:
            header += [self.obj.id_csv() + '_image']
        return header

    def export_data(self):
        data = {}
        if self.obj.questionnaire.csvdata.export_question_images:
            if self.obj.boxes:
                img = self.obj.questionnaire.csvdata.image_writer.output_boxes(self.obj.boxes, real=False)
                data[self.obj.id_csv() + '_image'] = img

        data[self.obj.id_csv() + '_review'] = str(self.obj.data.review_comment if self.obj.data.review_comment else '')

        return data

    def import_data(self, data):
        if self.obj.id_csv() + '_review' in data:
            self.obj.data.review_comment = data[self.obj.id_csv() + '_review']

class Choice(QObject, metaclass=model.buddy.Register):

    name = 'csvdata'
    obj_class = model.questionnaire.Choice

    def export_header(self):
        header = QObject.export_header(self)
        for box in self.obj.boxes:
            header += box.csvdata.export_header()

        return header

    def export_data(self):
        data = QObject.export_data(self)

        for box in self.obj.boxes:
            data.update(box.csvdata.export_data())

        return data

    def import_data(self, data):
        QObject.import_data(self, data)

        for box in self.obj.boxes:
            box.csvdata.import_data(data)

class Text(QObject, metaclass=model.buddy.Register):

    name = 'csvdata'
    obj_class = model.questionnaire.Text

    def export_header(self):
        header = QObject.export_header(self)
        for box in self.obj.boxes:
            header += box.csvdata.export_header()

        return header

    def export_data(self):
        data = QObject.export_data(self)

        for box in self.obj.boxes:
            data.update(box.csvdata.export_data())

        return data

    def import_data(self, data):
        QObject.import_data(self, data)

        for box in self.obj.boxes:
            box.csvdata.import_data(data)

class Option(QObject, metaclass=model.buddy.Register):

    name = 'csvdata'
    obj_class = model.questionnaire.Option

    def export_header(self):
        header = QObject.export_header(self)
        header += [self.obj.id_csv()]
        return header

    def export_data(self):
        data = {self.obj.id_csv(): '%i' % self.obj.get_answer()}
        data.update(QObject.export_data(self))
        return data

    def import_data(self, data):
        QObject.import_data(self, data)

        if self.obj.id_csv() in data:
            self.obj.set_answer(int(data[self.obj.id_csv()]))


class Additional_Mark(QObject, metaclass=model.buddy.Register):

    name = 'csvdata'
    obj_class = model.questionnaire.Additional_Mark

    def export_header(self):
        return QObject.export_header(self) + [self.obj.id_csv()]

    def export_data(self):
        data = {self.obj.id_csv(): '%i' % self.obj.get_answer()}
        data.update(QObject.export_data(self))
        return data

    def import_data(self, data):
        QObject.import_data(self, data)

        if self.obj.id_csv() in data:
            self.obj.set_answer(int(data[self.obj.id_csv()]))


class Box(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'csvdata'
    obj_class = model.questionnaire.Box

    def export_header(self):
        header = [self.obj.id_csv()]
        if self.obj.question.questionnaire.csvdata.export_quality:
            header += [self.obj.id_csv() + '_quality']
        return header

    def export_data(self):
        data = {self.obj.id_csv() : str(int(self.obj.data.state))}
        if self.obj.question.questionnaire.csvdata.export_quality:
            data.update({self.obj.id_csv() + '_quality' : self.obj.data.quality})
        return data

    def import_data(self, data):
        if not self.obj.id_csv() in data:
            return

        self.obj.data.state = int(data[self.obj.id_csv()])


class Textbox(Box, metaclass=model.buddy.Register):

    name = 'csvdata'
    obj_class = model.questionnaire.Textbox

    def export_header(self):
        header = [self.obj.id_csv()]
        return header

    def export_data(self):
        data = str(int(self.obj.data.state))

        image_writer = self.obj.question.questionnaire.csvdata.image_writer

        if self.obj.data.state and self.obj.data.text:
            data = self.obj.data.text
        elif self.obj.data.state and self.obj.question.questionnaire.csvdata.export_images:
            data = image_writer.output_box(self.obj)

        data = { self.obj.id_csv() : data }

        return data

    def import_data(self, data):
        if not self.obj.id_csv() in data:
            return

        try:
            state = int(data[self.obj.id_csv()])
            text = ''
        except ValueError:
            state = 1
            text = str(data[self.obj.id_csv()])

        self.obj.data.state = state
        self.obj.data.text = text

class Codebox(Textbox, metaclass=model.buddy.Register):

    name = 'csvdata'
    obj_class = model.questionnaire.Codebox

    def export_header(self):
        header = [self.obj.id_csv()]
        return header

    def export_data(self):
        data = str(int(self.obj.data.state))

        if self.obj.data.state and self.obj.data.text:
            data = self.obj.data.text

        data = { self.obj.id_csv() : data }

        return data

    def import_data(self, data):
        if not self.obj.id_csv() in data:
            return

        try:
            state = int(data[self.obj.id_csv()])
            text = ''
        except ValueError:
            state = 1
            text = str(data[self.obj.id_csv()])

        self.obj.data.state = state
        self.obj.data.text = text





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sdaps import model
from sdaps import log

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext


class QObject(model.buddy.Buddy, metaclass=model.buddy.Register):

    obj_class = model.questionnaire.QObject
    name = 'setup'

    def init(self):
        pass

    def question(self, chars):
        raise AssertionError('Setting a question on this QObject type is not possible.')

    def answer(self, chars):
        raise AssertionError('Adding an answer to this QObject type is not possible.')

    def box(self, box):
        raise AssertionError('Adding a box to this QObject type is not possible.')

    def variable_name(self, value):
        self.obj.var = value

    def validate(self):
        pass

    def setup(self, *args):
        pass


class Head(QObject, metaclass=model.buddy.Register):

    obj_class = model.questionnaire.Head

    def question(self, chars):
        self.obj.title += chars.strip()

    def validate(self):
        if not self.obj.title:
            log.warn(_('Head %(l0)i got no title.') % {'l0': self.obj.id[0]})


class Question(QObject, metaclass=model.buddy.Register):

    obj_class = model.questionnaire.Question

    def question(self, chars):
        self.obj.question += chars.strip()

    def validate(self):
        if not self.obj.question:
            log.warn(_('%(class)s %(l0)i.%(l1)i got no question.') % {
                'class': self.obj.__class__.__name__,
                'l0': self.obj.id[0], 'l1': self.obj.id[1]
            })


class Choice(Question, metaclass=model.buddy.Register):

    obj_class = model.questionnaire.Choice

    def init(self):
        self.box_cache = list()
        self.answer_cache = list()

    def _box(self, box):
        self.obj.add_box(box)
        if self.obj.page_number == 0:
            self.obj.page_number = box.page_number
        else:
            assert self.obj.page_number == box.page_number

    def answer(self, chars):
        self.answer_cache.append(chars)

    def box(self, box):
        self.box_cache.append(box)


    def setup(self):
        while self.box_cache and self.answer_cache:
            box = self.box_cache.pop(0)
            answer = self.answer_cache.pop(0)
            box.setup.answer(answer)
            self._box(box)

        Question.setup(self)

    def validate(self):
        Question.validate(self)
        if self.box_cache or self.answer_cache:
            raise AssertionError(_("Error in question \"%s\"") % self.obj.question)
        del self.box_cache
        del self.answer_cache
        if not self.obj.boxes:
            log.warn(_('%(class)s %(l0)i.%(l1)i got no boxes.') % {
                'class': self.obj.__class__.__name__,
                'l0': self.obj.id[0], 'l1': self.obj.id[1]
            })


class Option(Choice, metaclass=model.buddy.Register):

    obj_class = model.questionnaire.Option


class Range(Option, metaclass=model.buddy.Register):

    obj_class = model.questionnaire.Range

    def set_lower(self, box, answer):
        self.obj.answers = (answer, self.obj.answers[1])
        self.obj.range = (box, self.obj.range[1])

    def set_upper(self, box, answer):
        self.obj.answers = (self.obj.answers[0], answer)
        self.obj.range = (self.obj.range[0], box)

    def setup(self):
        # TODO: This happens while parsing the ODT, verify why and figure out
        #       whether this is the reasonable solution to it.
        if not self.box_cache:
            Option.setup(self)
            return

        # Insert None answer texts for the boxes part of the range
        range_len = self.obj.range[1] - self.obj.range[0] + 1

        self.answer_cache = \
            self.answer_cache[:self.obj.range[0]] + \
            [None for i in range(range_len)] + \
            self.answer_cache[self.obj.range[0]:]

        # Run normal setup routine which will match up the answers outside
        # the range correctly now.
        Option.setup(self)


    def validate(self):
        Option.validate(self)

        boxes = len(self.obj.boxes)

        if not (0 <= self.obj.range[0] < boxes):
            log.warn(_('%(class)s %(l0)i.%(l1)i lower box out of range.') % {
                'class': self.obj.__class__.__name__,
                'l0': self.obj.id[0], 'l1': self.obj.id[1]
            })
        if not (0 <= self.obj.range[1] < boxes):
            log.warn(_('%(class)s %(l0)i.%(l1)i upper box out of range.') % {
                'class': self.obj.__class__.__name__,
                'l0': self.obj.id[0], 'l1': self.obj.id[1]
            })
        if not (self.obj.range[0] < self.obj.range[1]):
            log.warn(_('%(class)s %(l0)i.%(l1)i lower box not before upper box.') % {
                'class': self.obj.__class__.__name__,
                'l0': self.obj.id[0], 'l1': self.obj.id[1]
            })

class Mark(Range, metaclass=model.buddy.Register):

    obj_class = model.questionnaire.Mark

    def init(self):
        Range.init(self)
        self.answer_count = 0

    def answer(self, chars):
        if self.answer_count == 0:
            self.obj.answers = (chars, '')
        elif self.answer_count == 1:
            self.obj.answers = (self.obj.answers[0], chars)

        self.answer_count += 1

    def box(self, box):
        assert isinstance(box, model.questionnaire.Checkbox)
        self.obj.add_box(box)
        box.value = len(self.obj.boxes)

        if self.obj.page_number == 0:
            self.obj.page_number = box.page_number
        else:
            assert self.obj.page_number == box.page_number

        self.obj.range = (0, len(self.obj.boxes) - 1)

    def validate(self):
        Range.validate(self)
        if self.answer_count != 2:
            log.warn(_('%(class)s %(l0)i.%(l1)i got not exactly two answers.') % {
                'class': self.obj.__class__.__name__,
                'l0': self.obj.id[0], 'l1': self.obj.id[1]
            })


class Text(Question, metaclass=model.buddy.Register):

    obj_class = model.questionnaire.Text

    def box(self, box):
        assert isinstance(box, model.questionnaire.Textbox)
        self.obj.add_box(box)
        if self.obj.page_number == 0:
            self.obj.page_number = box.page_number
        else:
            assert self.obj.page_number == box.page_number

    def validate(self):
        Question.validate(self)
        if not len(self.obj.boxes) == 1:
            log.warn(_('%(class)s %(l0)i.%(l1)i got not exactly one box.') % {
                'class': self.obj.__class__.__name__,
                'l0': self.obj.id[0], 'l1': self.obj.id[1]
            })


class Additional_Head(Head, metaclass=model.buddy.Register):

    obj_class = model.questionnaire.Additional_Head

    def setup(self, args):
        assert len(args) == 1
        self.question(args[0])
        self.validate()


class Additional_Mark(Question, metaclass=model.buddy.Register):

    obj_class = model.questionnaire.Additional_Mark

    def setup(self, args):
        assert len(args) == 3
        self.question(args[0])
        self.obj.answers.append(args[1])
        self.obj.answers.append(args[2])
        self.validate()


class Additional_FilterHistogram(Question, metaclass=model.buddy.Register):

    obj_class = model.questionnaire.Additional_FilterHistogram

    def setup(self, args):
        assert len(args) % 2 == 1
        self.question(args.pop(0))
        while len(args):
            self.obj.answers.append(args.pop(0))
            self.obj.filters.append(args.pop(0))
        self.validate()


class Box(model.buddy.Buddy, metaclass=model.buddy.Register):

    obj_class = model.questionnaire.Box
    name = 'setup'

    def setup(self, page_number, x, y, width, height, lw=None):
        self.obj.page_number = page_number
        self.obj.x = x
        self.obj.y = y
        self.obj.width = width
        self.obj.height = height
        if lw is not None:
            self.obj.lw = lw

    def answer(self, text):
        self.obj.text = text




#!/usr/bin/env python3
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2011, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This file generates the latexmap.py file which is needed
to convert latex symbols into their unicode counterpart.

Run with create-latexmap.py path/to/utf8enc.dfu
"""

import re
import sys

if len(sys.argv) != 2:
    print(__doc__)
    sys.exit(1)

output = open('latexmap.py', 'w')
input = open(sys.argv[1])

data = input.read()

regexp = re.compile(r'^\\DeclareUnicodeCharacter\{(?P<unicode>[0-9a-fA-F]{4})\}\{(?P<str>[^\}]+)\}', re.MULTILINE)

mapping = []
for match in regexp.finditer(data):
    repr = match.group('str')
    repr = repr.replace('\\@tabacckludge', '\\')
    repr = repr.replace('\\', '\\\\').replace('\'', '\\\'')
    mapping.append('''    u\'%s\': u\'\\u%s\'''' %
                   (repr,
                    match.group('unicode')))

output.write('''#This file is auto generated from the latex unicode mapping.

#: Mapping from LaTeX commands to unicode characters.
mapping = {
''')

output.write(',\n'.join(mapping))
output.write('''
}
''')



# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2012, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module reset already stored data so next reports and exports will contain only new questionnaires.
"""

from collections import defaultdict
from sdaps import model
import bz2,os,pickle
from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

def reset(survey):
    print((_("Removing stored data...")))
    survey.iterate(lambda: survey.delete_sheet(survey.get_sheet()))
    survey.questionnaire_ids = []
    survey.save()
    print((_("Done")))



# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sdaps import model
from sdaps import script
from sdaps import log

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext


parser = script.add_project_subparser("info",
    help=_("Display and modify metadata of project."),
    description=_("""This command lets you modify the metadata of the SDAPS
    project. You can modify, add and remove arbitrary keys that will be printed
    on the report. The only key that always exist is "title".
    If no key is given then a list of defined keys is printed."""))

parser.add_argument('-d', '--delete',
    action="store_true",
    help=_("Delete the key and value pair."))

parser.add_argument('key',
    nargs="?",
    help=_("The key to display or modify."))

parser.add_argument('value',
    nargs="?",
    help=_("Set the given key to this value."))



@script.connect(parser)
@script.logfile
def info(cmdline):
    survey = model.survey.Survey.load(cmdline['project'])

    if cmdline['key']:
        key = cmdline['key'].strip()
        if cmdline['value']:
            value = cmdline['value'].strip()
            if key == "title":
                survey.title = value
            else:
                survey.info[key] = value
        elif cmdline['delete']:
            del survey.info[key]
        else:
            if key == "title":
                print(survey.title)
            else:
                print(survey.info[key])
    else:
        log.interactive(_('Existing fields:\n'))
        print("title")
        for key in list(survey.info.keys()):
            print(key)

    survey.save()





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sdaps import log
import xml.sax
import zipfile
import re

from sdaps import model
from sdaps.utils.latex import latex_to_unicode

QOBJECT_PREFIX = 'QObject'
ANSWER_PREFIX = 'Answer'
BOX = 'Box'
VARIABLE = 'Variable'
TEXTBOX = 'Textbox'
RANGE_PREFIX = 'Range'

index_re = re.compile(r'''^(?P<index>(?:[0-9]+\.)+)(?P<string>.*)$''')
arg_index_re = re.compile(r'''^(?P<arg>[^\[]*)(\[(?P<index>([0-9]+\.)*[0-9]+)\])?$''')


def get_index_and_string(string):
    match = index_re.match(string)
    if match is None:
        if string.startswith('XAUTO. '):
            return None, string[7:]
        return None, string

    string = match.group('string')
    index = match.group('index')
    index = index.split('.')[:-1]
    index = tuple([int(x) for x in index])

    return index, string

def parse(survey):

    sdaps_file = open(survey.path('questionnaire.sdaps'))
    # the file is encoded in ascii format
    sdaps_data = sdaps_file.read()
    qobject = None
    auto_numbering_id = (0,)

    sdaps_data = sdaps_data.split('\n')

    if sdaps_data[0].startswith('['):
        lines = []
        for line in sdaps_data:
            # Ignore empty lines
            if not line:
                continue
            num, line = line.split(']', 1)
            num = int(num[1:])
            lines.append((num, line))

        lines.sort(key=lambda x: x[0])

        sdaps_data = [l[1] for l in lines]

    for line in sdaps_data:
        line = line.strip()
        if line == "":
            continue
        arg, value = line.split('=', 1)
        arg = arg.strip()
        value = value.strip()
        value = latex_to_unicode(value)

        match = arg_index_re.match(arg)
        if match is not None:
            arg = match.group('arg')
            index = match.group('index')
            if index is None:
                try:
                    survey.questionnaire.qobjects[-1]
                except IndexError:
                    pass
            else:
                index = tuple([int(s) for s in index.split('.')])
                qobject = survey.questionnaire.find_object(index)


        if arg == 'Title':
            survey.title = value
        elif arg == 'PrintQuestionnaireId':
            survey.defs.print_questionnaire_id = bool(int(value))
        elif arg == 'PrintSurveyId':
            survey.defs.print_survey_id = bool(int(value))
        elif arg == 'Pages':
            survey.questionnaire.page_count = int(value)
        elif arg == 'CheckMode':
            survey.defs.checkmode = value
            assert survey.defs.checkmode in model.survey.valid_checkmodes
        elif arg == 'GlobalID':
            survey.global_id = value
        elif arg == 'GlobalIDLabel':
            # Ignore for now
            pass
        elif arg == 'Duplex':
            survey.defs.duplex = (value.lower() == "true")
        elif arg == 'Style':
            survey.defs.style = value
            assert survey.defs.style in model.survey.valid_styles
        elif arg == "PageSize":
            args = value.split(',')
            args = [arg.strip() for arg in args]

            width, height = [round(float(arg[:-2]) / 72.27 * 25.4, 3) for arg in args]

            survey.defs.paper_width = width
            survey.defs.paper_height = height

        elif arg.startswith(QOBJECT_PREFIX):
            index, string = get_index_and_string(value)
            if index:
                auto_numbering_id = index + (0,)
            else:
                auto_numbering_id = auto_numbering_id[:-1] + (auto_numbering_id[-1] + 1,)
                index = auto_numbering_id

            qobject_type = arg[len(QOBJECT_PREFIX) + 1:]

            qobject_type = qobject_type.lower().capitalize()

            qobject = getattr(model.questionnaire, qobject_type)
            assert issubclass(qobject, model.questionnaire.QObject)
            qobject = qobject()
            survey.questionnaire.add_qobject(qobject, new_id=index)
            qobject.setup.init()

            qobject.setup.question(string)
        elif arg == VARIABLE:
            assert qobject is not None

            qobject.setup.variable_name(value)

        elif arg.startswith(ANSWER_PREFIX):
            assert qobject is not None

            answer_type = arg[len(ANSWER_PREFIX) + 1:]

            qobject.setup.answer(value)

        elif arg.startswith(RANGE_PREFIX):
            assert qobject is not None
            assert isinstance(qobject, model.questionnaire.Range)

            idx, answer = value.split(',', 1)
            idx = int(idx)

            range_type = arg[len(RANGE_PREFIX) + 1:].lower()

            if range_type == 'lower':
                qobject.setup.set_lower(idx, answer)
            elif range_type == 'upper':
                qobject.setup.set_upper(idx, answer)
            else:
                raise AssertionError('File format error, %s has to be either lower or upper!' % RANGE_PREFIX)

        elif arg == BOX:
            args = value.split(',')
            args = [arg.strip() for arg in args]

            boxtype = args[0]
            # Convert to mm
            page = int(args[1])
            x, y, width, height = [float(arg[:-2]) / 72.27 * 25.4 for arg in args[2:6]]
            y = survey.defs.paper_height - y
            lw = None

            if boxtype == 'Textbox':
                box = model.questionnaire.Textbox()
                if len(args) == 9:
                    lw = args[6] if args[6] else None
                    box.var = args[7] if args[7] else None
                    box.value = int(args[8]) if args[8] else None
                else:
                    assert(len(args) == 6)
            elif boxtype == 'Codebox':
                box = model.questionnaire.Codebox()
                if len(args) == 9:
                    lw = args[6] if args[6] else None
                    box.var = args[7] if args[7] else None
                    box.value = int(args[8]) if args[8] else None
                else:
                    assert(len(args) == 6)
            else:
                box = model.questionnaire.Checkbox()
                if len(args) == 7:
                    box.form = args[6]
                elif len(args) == 10:
                    box.form = args[6]
                    lw = args[7] if args[7] else None
                    box.var = args[8] if args[8] else None
                    box.value = int(args[9]) if args[9] else None
                else:
                    assert(len(args) == 6)

            if lw:
                lw = float(lw[:-2]) / 72.27 * 25.4

            box.setup.setup(page, x, y, width, height, lw)
            qobject.setup.box(box)
        else:
            # Falltrough, it is some metadata:
            survey.info[arg] = value

    # Force duplex of for one page questionnaires
    if survey.questionnaire.page_count == 1:
        survey.defs.duplex = False




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2018, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from types import ModuleType


def toJson(obj):

    if hasattr(obj, '__slots__'):
        res = dict()
        for slot in getattr(obj, '__slots__'):
            res[slot] = getattr(obj, slot)

    else:
        if hasattr(obj, '_save_attrs'):
            attrs = getattr(obj, '_save_attrs')
            res = dict()
            for attr in attrs:
                res[attr] = getattr(obj, attr)
        else:
            if hasattr(obj, '_save_skip'):
                skip = getattr(obj, '_save_skip')
            else:
                skip = set()

            res = obj.__dict__.copy()
            keys = list(res.keys())
            for key in keys:
               if key.startswith('_') or key in skip:
                   del res[key]

    if hasattr(obj, '__to_json_state__'):
        obj.__to_json_state__(res)

    res['_class'] = obj.__class__.__name__
    return res

def fromJson(data, module_or_class):
    if isinstance(module_or_class, ModuleType):
        cls = getattr(module_or_class, data['_class'])
    else:
        cls = module_or_class

    obj = cls.__new__(cls)
    del data['_class']

    if hasattr(cls, '__setstate__'):
        getattr(cls, '__setstate__')
        cls.__setstate__(obj, data)

    elif hasattr(obj, '__slots__'):
        for k, v in data.items():
            setattr(obj, k, v)

    else:
        obj.__dict__.update(data)

    return obj





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, 2013, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# This file duplicates some drawing code from the gui module.

from gi.repository import Pango
from gi.repository import PangoCairo
import cairo
import math

from sdaps import model
from sdaps import defs

colorcycle_pos = 0
colorcycle = [(1, 0, 0), (0.5, 0.5, 0), (0, 0.5, 0.5), (0.5, 0, 0.5)]

LINE_WIDTH = 25.4/72
MIN_FREETEXT_SIZE = 4.0

def cycle_color():
    global colorcycle_pos
    colorcycle_pos = (colorcycle_pos + 1) % len(colorcycle)

def set_rgb_from_color_cycle(cr):
    global colorcycle_pos, colorcycle
    cr.set_source_rgba(*(colorcycle[colorcycle_pos] + (0.5,)))

def inner_box(cr, x, y, width, height):
    line_width = cr.get_line_width()

    cr.rectangle(x + line_width / 2.0, y + line_width / 2.0,
                 width - line_width, height - line_width)

def inner_ellipse(cr, x, y, width, height):
    cr.save()

    cr.translate(x + width / 2.0, y + height / 2.0)

    line_width = cr.get_line_width()

    cr.scale((width - line_width) / 2.0, (height - line_width) / 2.0)
    cr.arc(0, 0, 1.0, 0, 2*math.pi)

    # Restore old matrix (without removing the current path)
    cr.restore()

def create_layout(cr, text, layout_info, indent=0):
    layout = PangoCairo.create_layout(cr)
    layout.set_text(text, -1)
    # Dont recreate the description all the time?
    font = Pango.FontDescription(layout_info['font'])
    layout.set_font_description(font)
    layout.set_width((layout_info['twidth'] - 2 * indent) * Pango.SCALE)
    layout.set_wrap(Pango.WrapMode.WORD_CHAR)

    layout.indent = indent

    return layout

def show_layout(cr, layout, layout_info, skip=1.25):
    x, y = layout_info['xshift'] + layout.indent, layout_info['ypos'] + skip
    cr.move_to(x, y)

    cr.set_source_rgb(0, 0, 0)
    PangoCairo.show_layout(cr, layout)

    layout_info['ypos'] = layout_info['ypos'] + layout.get_pixel_size()[1] + skip

    # Return the position of the baseline
    y += layout.get_baseline() / Pango.SCALE

    return x, y

class Questionnaire(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'annotate'
    obj_class = model.questionnaire.Questionnaire

    def draw(self, cr, page_number, layout_info):
        cr.set_source_rgba(0.0, 0.0, 1.0, 0.5)
        cr.set_line_width(LINE_WIDTH)

        # Draw corner marks.
        cr.move_to(defs.corner_mark_left + defs.corner_mark_length, defs.corner_mark_top)
        cr.line_to(defs.corner_mark_left, defs.corner_mark_top)
        cr.line_to(defs.corner_mark_left, defs.corner_mark_top+defs.corner_mark_length)

        cr.move_to(self.obj.survey.defs.paper_width - defs.corner_mark_right - defs.corner_mark_length, defs.corner_mark_top)
        cr.line_to(self.obj.survey.defs.paper_width - defs.corner_mark_right, defs.corner_mark_top)
        cr.line_to(self.obj.survey.defs.paper_width - defs.corner_mark_right, defs.corner_mark_top + defs.corner_mark_length)

        cr.move_to(defs.corner_mark_left + defs.corner_mark_length, self.obj.survey.defs.paper_height - defs.corner_mark_bottom)
        cr.line_to(defs.corner_mark_left, self.obj.survey.defs.paper_height - defs.corner_mark_top)
        cr.line_to(defs.corner_mark_left, self.obj.survey.defs.paper_height - defs.corner_mark_top - defs.corner_mark_length)

        cr.move_to(self.obj.survey.defs.paper_width - defs.corner_mark_right - defs.corner_mark_length, self.obj.survey.defs.paper_height - defs.corner_mark_top)
        cr.line_to(self.obj.survey.defs.paper_width - defs.corner_mark_right, self.obj.survey.defs.paper_height - defs.corner_mark_top)
        cr.line_to(self.obj.survey.defs.paper_width - defs.corner_mark_right, self.obj.survey.defs.paper_height - defs.corner_mark_top - defs.corner_mark_length)

        cr.stroke()

        for qobject in self.obj.qobjects:
            cycle_color()
            qobject.annotate.draw(cr, page_number, layout_info)


class QObject(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'annotate'
    obj_class = model.questionnaire.QObject

    def draw(self, cr, page_number, layout_info):
        pass


class Head(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'annotate'
    obj_class = model.questionnaire.Head

    def draw(self, cr, page_number, layout_info):
        # There is no page number for head objects.
        layout = create_layout(cr, self.obj.id_str() + " " + self.obj.title, layout_info)
        xpos, ypos = show_layout(cr, layout, layout_info)


class Question(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'annotate'
    obj_class = model.questionnaire.Question

    def draw(self, cr, page_number, layout_info):
        # Does the sheet contain this question?
        if page_number == self.obj.page_number:
            layout = create_layout(cr, self.obj.id_str() + " " + self.obj.question, layout_info, 3)
            xpos, ypos = show_layout(cr, layout, layout_info, 3)

            # iterate over boxes
            for box in self.obj.boxes:
                box.annotate.draw(cr, layout_info)

class Range(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'annotate'
    obj_class = model.questionnaire.Range

    def draw(self, cr, page_number, layout_info):
        # Does the sheet contain this question?
        if page_number == self.obj.page_number:
            layout = create_layout(cr, self.obj.id_str() + " " + self.obj.question, layout_info, 3)
            xpos, ypos = show_layout(cr, layout, layout_info, 3)

            # Lower range
            layout = create_layout(cr, self.obj.answers[0], layout_info, 10)
            xpos, ypos = show_layout(cr, layout, layout_info, 3)

            cr.move_to(self.obj.boxes[self.obj.range[0]].x, self.obj.boxes[self.obj.range[0]].y)
            cr.line_to(xpos, ypos)
            cr.set_line_width(LINE_WIDTH)
            set_rgb_from_color_cycle(cr)
            cr.stroke()

            # Upper range
            layout = create_layout(cr, self.obj.answers[1], layout_info, 10)
            xpos, ypos = show_layout(cr, layout, layout_info, 3)

            cr.move_to(self.obj.boxes[self.obj.range[1]].x, self.obj.boxes[self.obj.range[1]].y)
            cr.line_to(xpos, ypos)
            cr.set_line_width(LINE_WIDTH)
            set_rgb_from_color_cycle(cr)
            cr.stroke()

            for i, box in enumerate(self.obj.boxes):
                hide_if_empty = (self.obj.range[0] <= i <= self.obj.range[1])

                box.annotate.draw(cr, layout_info, hide_if_empty=hide_if_empty)

class Box(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'annotate'
    obj_class = model.questionnaire.Checkbox

    def draw_box(self, cr):
        inner_box(cr, self.obj.x, self.obj.y, self.obj.width, self.obj.height)
        cr.stroke()

    def draw(self, cr, layout_info, hide_if_empty=False):
        if not hide_if_empty or self.obj.text:
            layout = create_layout(cr, self.obj.id_str() + " " + self.obj.text, layout_info, 6)
            xpos, ypos = show_layout(cr, layout, layout_info)

            cr.move_to(self.obj.x, self.obj.y)
            cr.line_to(xpos, ypos)
            cr.set_line_width(LINE_WIDTH)
            set_rgb_from_color_cycle(cr)
            cr.stroke()

        cr.set_fill_rule(cairo.FILL_RULE_EVEN_ODD)

        cr.set_source_rgba(0.0, 0.0, 1.0, 0.5)
        cr.set_line_width(LINE_WIDTH)

        self.draw_box(cr)

        text = str(self.obj.id[-1])
        layout = PangoCairo.create_layout(cr)
        layout.set_text(text, len(text))
        # Dont recreate the description all the time?
        font = Pango.FontDescription(layout_info['boxfont'])
        layout.set_font_description(font)

        cr.move_to(self.obj.x + LINE_WIDTH, self.obj.y + LINE_WIDTH)
        PangoCairo.show_layout(cr, layout)
        cr.new_path()


class Checkbox(Box, metaclass=model.buddy.Register):

    name = 'annotate'
    obj_class = model.questionnaire.Checkbox

    def draw_box(self, cr):
        if self.obj.form == "box":
            inner_box(cr, self.obj.x, self.obj.y, self.obj.width, self.obj.height)
            cr.stroke()
        elif self.obj.form == "ellipse":
            inner_ellipse(cr, self.obj.x, self.obj.y, self.obj.width, self.obj.height)
            cr.stroke()



class Textbox(Box, metaclass=model.buddy.Register):

    name = 'annotate'
    obj_class = model.questionnaire.Textbox





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# The code uses multiple inheritance; however, most of the additional base
# classes are solely used as mixins. This means they must not contain any
# __init__ functions, as that could cause trouble.


from . import db
from . import buddy
from . import data
import sys
import struct


class DataObject(object):
    '''Mixin
    '''

    def get_data(self):
        if not self.id in self.sheet.data:
            if hasattr(self, '_data_object'):
                clsname = self._data_object
            else:
                clsname = self.__class__.__name__
            self.sheet.data[self.id] = getattr(data, clsname)(self)
        return self.sheet.data[self.id]

    data = property(get_data)


class Questionnaire(buddy.Object):
    '''
    Identification: There is only one.
    Reference: survey.questionnaire
    Parent: self.survey
    '''

    _save_skip = {'survey'}

    def __init__(self):
        self.survey = None
        self.qobjects = list()
        self.last_id = (-1,)
        self.init_attributes()
        self._notify_changed_list = list()

    def init_attributes(self):
        self.page_count = 0

    def add_qobject(self, qobject, new_id=None):
        qobject.questionnaire = self
        # XXX: Is this any good?
        if new_id is not None:
            assert new_id > self.last_id
            self.last_id = new_id
            qobject.id = new_id
        else:
            self.last_id = qobject.init_id(self.last_id)
        self.qobjects.append(qobject)

    def get_sheet(self):
        return self.survey.sheet

    sheet = property(get_sheet)

    def notify_data_changed(self, qobj, dobj, name, old_value):
        for func in self._notify_changed_list:
            func(self, qobj, dobj, name, old_value)

    def connect_data_changed(self, func):
        self._notify_changed_list.append(func)

    def disconnect_data_changed(self, func):
        self._notify_changed_list.remove(func)

    def __str__(self):
        return str().join(
            ['%s\n' % self.__class__.__name__] +
            [str(qobject) for qobject in self.qobjects]
        )

    def find_object(self, oid):
        for qobject in self.qobjects:
            obj = qobject.find_object(oid)

            if obj is not None:
                return obj

    def reinit_state(self):
        self._notify_changed_list = list()

    def __setstate__(self, data):
        self.__dict__.update(data)

        self._notify_changed_list = list()

        for i in range(len(self.qobjects)):
            self.qobjects[i] = db.fromJson(self.qobjects[i], sys.modules[__name__])
            self.qobjects[i].questionnaire = self

class QObject(buddy.Object, DataObject):
    '''
    Identification: id ==(major, minor)
    Reference: survey.questionnaire.qobjects[i](i != id)
    Parent: self.questionnaire
    '''

    _data_object = 'QObject'
    _save_skip = {'survey', 'questionnaire'}

    def __init__(self):
        self.questionnaire = None
        self.boxes = list()
        self.last_id = -1
        self.max_value = -1
        self.var = None
        self.init_attributes()

    def init_attributes(self):
        pass

    def init_id(self, id):
        self.id = id[:-1] + (id[-1] + 1,)
        return self.id

    def add_box(self, box):
        box.question = self
        self.last_id = box.init_id(self.last_id)
        self.max_value = max(self.max_value, box.init_value(self.max_value))
        self.boxes.append(box)

    def get_sheet(self):
        return self.questionnaire.sheet

    sheet = property(get_sheet)

    def calculate_survey_id(self, md5):
        pass

    def id_str(self):
        ids = [str(x) for x in self.id]
        return '.'.join(ids)

    def id_csv(self):
        if self.var:
            return self.var

        ids = [str(x) for x in self.id]
        return '_'.join(ids)

    def id_filter(self):
        ids = [str(x) for x in self.id]
        return '_' + '_'.join(ids)

    def __str__(self):
        return '(%s)\n' % (
            self.__class__.__name__,
        )

    def find_object(self, oid):
        if self.id == oid:
            return self

        for box in self.boxes:
            obj = box.find_object(oid)

            if obj is not None:
                return obj

        return None

    def __setstate__(self, data):
        # Attributes that may not (yet) be present in the database
        self.var = None

        self.__dict__.update(data)
        self.id = tuple(self.id)

        for i in range(len(self.boxes)):
            self.boxes[i] = db.fromJson(self.boxes[i], sys.modules[__name__])
            self.boxes[i].question = self

class Head(QObject):

    def init_attributes(self):
        QObject.init_attributes(self)
        self.title = str()

    def init_id(self, id):
        self.id = (id[0] + 1, ) + (0,)*(len(id)-1)
        return self.id

    def __str__(self):
        return '%s(%s) %s\n' % (
            self.id_str(),
            self.__class__.__name__,
            self.title,
        )


class Question(QObject):

    def init_attributes(self):
        QObject.init_attributes(self)
        self.page_number = 0
        self.question = str()

    def calculate_survey_id(self, md5):
        for box in self.boxes:
            box.calculate_survey_id(md5)

    def __str__(self):
        return '%s(%s) %s {%i}\n' % (
            self.id_str(),
            self.__class__.__name__,
            self.question,
            self.page_number
        )


class Choice(Question):

    def __str__(self):
        return str().join(
            [Question.__str__(self)] +
            [str(box) for box in self.boxes]
        )

    def get_answer(self):
        '''it's a list containing all selected values
        '''
        answer = list()
        for box in self.boxes:
            if box.data.state:
                answer.append(box.value)
        return answer

class Option(Question):

    def init_attributes(self):
        Question.init_attributes(self)
        self.value_none = -1
        self.value_invalid = -2

    def __str__(self):
        return str().join(
            [Question.__str__(self)] +
            [str(box) for box in self.boxes]
        )

    def get_answer(self):
        answer = list()
        for box in self.boxes:
            if box.data.state:
                answer.append(box.value)

        if len(answer) == 1:
            return answer[0]
        else:
            if len(answer) == 0:
                return self.value_none
            else:
                return self.value_invalid

    def set_answer(self, answer):
        # Used for CSV import. Set the first box with the right value, unset
        # all others.
        # Raises an exception if not found

        # Unset all if value is the none value
        if answer == self.value_none:
            for box in self.boxes:
                box.data.state = 0
            return

        # Set all if value is the none value
        # NOTE: Assumes more than one box!
        if answer == self.value_invalid:
            assert len(self.boxes) > 1
            for box in self.boxes:
                box.data.state = 1
            return

        found = False
        for box in self.boxes:
            if not found and box.value == answer:
                found = True
                box.data.state = 1
            else:
                box.data.state = 0

        assert(found)

class Range(Option):

    def init_attributes(self):
        Option.init_attributes(self)
        self.answers = ("", "")
        self.range = (0, 0)

    def __str__(self):
        if len(self.answers) == 2:
            return str().join(
                [Question.__str__(self)] +
                ['\t%s (%i) - %s (%i)\n' % (self.answers[0], self.range[0], self.answers[1], self.range[1])] +
                [str(box) for box in self.boxes]
            )
        else:
            return str().join(
                [Question.__str__(self)] +
                ['\t? - ?\n'] +
                [str(box) for box in self.boxes]
            )

class Mark(Range):
    # Just an alias for unpickling old data
    pass

class Text(Question):

    def __str__(self):
        return str().join(
            [Question.__str__(self)] +
            [str(box) for box in self.boxes]
        )

    def get_answer(self):
        '''it's a bool, wether there is content in the textbox
        '''
        assert len(self.boxes) == 1
        text = ""
        for box in self.boxes:
            text = text + box.data.text
        if text:
            return text
        else:
            return self.boxes[0].data.state


class Additional_Head(Head):

    pass


class Additional_Mark(Question):

    def init_attributes(self):
        Question.init_attributes(self)
        self.answers = list()

    def __str__(self):
        return str().join(
            [Question.__str__(self)] +
            ['\t%s - %s\n' % tuple(self.answers)]
        )

    def get_answer(self):
        return self.data.value

    def set_answer(self, answer):
        self.data.value = answer


class Additional_FilterHistogram(Question):

    def init_attributes(self):
        Question.init_attributes(self)
        self.answers = list()
        self.filters = list()

    def __str__(self):
        result = []
        result.append(Question.__str__(self))
        for i in range(len(self.answers)):
            result.append('\t%s - %s\n' % (self.answers[i], self.filters[i]))
        return str().join(result)

    def get_answer(self):
        return self.data.value

    def set_answer(self, answer):
        raise NotImplemented()


class Box(buddy.Object, DataObject):
    '''
    Identification: id of the parent and value of the box ::

        id == (major, minor, value)

    Reference: survey.questionnaire.qobjects[i].boxes[j]
    Parent: self.question
    '''

    _save_skip = { 'question' }

    def __init__(self):
        self.question = None
        self.init_attributes()

    def init_attributes(self):
        self.page_number = 0
        self.x = 0
        self.y = 0
        self.width = 0
        self.height = 0
        self.lw = 25.4 / 72.0
        self.text = str()

        self.var = None
        self.value = None

    def init_id(self, id):
        id = id + 1

        self.id = self.question.id + (id,)
        return id

    def init_value(self, value):
        value = value + 1

        if self.value is None:
            self.value = value

        return self.value

    def id_str(self):
        ids = [str(x) for x in self.id]
        return '.'.join(ids)

    def id_csv(self):
        if self.var:
            return self.var

        ids = [str(x) for x in self.id]
        return '_'.join(ids)

    def get_sheet(self):
        return self.question.sheet

    sheet = property(get_sheet)

    def calculate_survey_id(self, md5):
        tmp = struct.pack('!ffff', self.x, self.y, self.width, self.height)
        md5.update(tmp)

    def __str__(self):
        return '\t%i(%s) %s %s %s %s %s\n' % (
            self.value,
            (self.__class__.__name__).ljust(8),
            ('%.1f' % self.x).rjust(5),
            ('%.1f' % self.y).rjust(5),
            ('%.1f' % self.width).rjust(5),
            ('%.1f' % self.height).rjust(5),
            self.text
        )

    def find_object(self, oid):
        if self.id == oid:
            return self

    def __setstate__(self, data):
        self.__dict__.update(data)
        self.id = tuple(self.id)

class Checkbox(Box):

    def init_attributes(self):
        Box.init_attributes(self)
        self.form = "box"

    def calculate_survey_id(self, md5):
        Box.calculate_survey_id(self, md5)
        md5.update(self.form.encode('utf-8'))

class Textbox(Box):

    pass

class Codebox(Textbox):

    _data_object = 'Textbox'




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Defs
====

This module contains constants and some magic values.
"""


# Corner Marks ============================================

# The position of the corner markings from the side of the paper
corner_mark_left = 10.0 # mm
corner_mark_right = 10.0 # mm
corner_mark_top = 12.0 # mm
corner_mark_bottom = 12.0 # mm
corner_mark_length = 20.0 # mm

# Bottom page margin (for QR-Code size calculation in fallback rendering)
bottom_page_margin = 25.0 # mm

# Length in mm of the corner marks in the scanned image
corner_mark_min_length = 15 # mm
corner_mark_max_length = 22 # mm

# The distance into the image that will be searched for the corner mark.
# Basically a square area of this size will be searched in each corner.
corner_mark_search_distance = 50 # mm

# Corner Boxes ============================================

# What corners are filled for each page, this is choosen so that
# the page number and rotation can be derived from the information.
# Order: top left, top right, bottom left, bottom right
corner_boxes = [
    [0, 1, 1, 1],
    [1, 1, 0, 0],
    [1, 0, 1, 1],
    [1, 0, 1, 0],
    [1, 0, 0, 0],
    [0, 0, 0, 1],
]

# Size of the corner boxes(measured from the center of the line)
corner_box_width = 3.5 # mm
corner_box_height = 3.5 # mm
# Distance from the corner marker(and also the distance to the codebox)
corner_box_padding = 3 # mm

# The coverage above which a cornerbox is considered to be a logical 1
cornerbox_on_coverage = 0.6


# Codebox =================================================

# Length in bits
codebox_length = 16 # bits
codebox_step = 3.5 # mm
codebox_width = codebox_step * codebox_length # mm
codebox_height = 3.5 # mm
# Padding that is ignored when testing the image whether it is 1/0
codebox_offset = 0.75 # mm

# The font and size of the font for the text that is printed between the codeboxes
codebox_text_font = 'Courier-Bold'
codebox_text_font_size = 10.5 # pt
# This is added to the vertical position to shift the text visually into the center
codebox_text_baseline_shift = 3 # mm

# Magic values for recognition
# The coverage above which a codebox is considered to be a logical 1
codebox_on_coverage = 0.7

# Code 128 Barcodes =======================================

code128_barwidth = 0.33 # mm. This is 0.93pt or 3.89px after scanning
code128_height = 6.5 # pt(5 mm)
code128_hpad = 6.5 # mm
code128_vpad = 4.02 # mm

code128_text_font = 'Courier'
code128_text_font_size = 9 # pt

# Checkbox ================================================

checkbox_metrics = {}

# The metrics is a mapping from the value to the quality and expected
# checkbox state.
# It works by searching the interval that we are in, and doing
# a linear interpolation.
# At state changes the point should be inserted twice.
# The touple is(metric, state, quality). To disable one you can just
# insert two dummy points with zero quality. To try and find better
# values have a look at the output of "boxgallery". Any suggestions
# for improvements(also algorithmic wise) are always welcome!
checkbox_metrics['checkcorrect'] = {}
checkbox_metrics['checkcorrect']['coverage'] = \
    [(0, 0, 1.0), (0.02, 0, 0.9), (0.05, 0, 0.3), (0.05, 1, 0.3),
     (0.1, 1, 1.0), (0.4, 1, 1.0), (0.5, 1, 0.2), (0.5, 0, 0.2),
     (0.7, 0, 0.3), (1.0, 0, 0.6)]
checkbox_metrics['checkcorrect']['cov-lines-removed'] = \
    [(0, 1, 0), (0.01, 1, 0), (0.07, 1, 1.0), (0.10, 1, 1.0),
     (0.13, 1, 0.3), (0.13, 0, 0.3), (0.25, 0, 0.7), (1, 0, 0.7)]
checkbox_metrics['checkcorrect']['cov-min-size'] = \
    [(0, 0, .9), (0.35, 0, 0.0), (0.35, 1, 0.0), (0.5, 1, 0.9),
     (0.55, 1, 1.0), (0.8, 1, 1.0), (0.9, 1, 0.9), (0.95, 1, 0.5),
     (0.95, 0, 0.5), (0.99, 0, 0.9), (1.0, 0, 1.0)]

# Same as above, but no correction mode. A light check is enough.
checkbox_metrics['check'] = {}
checkbox_metrics['check']['coverage'] = \
    [(0, 0, 1.0), (0.02, 0, 0.9), (0.05, 0, 0.3), (0.05, 1, 0.3),
     (0.1, 1, 1.0), (1.0, 0, 1.0)]
checkbox_metrics['check']['cov-lines-removed'] = \
    [(0, 1, 0), (1, 1, 0)]
checkbox_metrics['check']['cov-min-size'] = \
    [(0, 1, 0), (1, 1, 0)]

# Similar, but a light checkmark, or just dirt will be ignored.
checkbox_metrics['fill'] = {}
checkbox_metrics['fill']['coverage'] = \
    [(0, 0, 1.0), (0.2, 0, 0.9), (0.3, 0, 0), (0.3, 1, 0),
     (0.4, 1, 1.0), (1.0, 1, 1.0)]
checkbox_metrics['fill']['cov-lines-removed'] = \
    [(0, 1, 0), (1, 1, 0)]
checkbox_metrics['fill']['cov-min-size'] = \
    [(0, 1, 0), (1, 1, 0)]


# Textbox =================================================

# Tolerance for the text box corner adjustment. If the difference is more than
# this in mm, the adjusted values will be discarded.
find_box_corners_tolerance = 1.5 # mm

# The size in mm of the area that is scanned during checking of textboxes. The step
# is the x/y distance in mm of each checked area. They overlap on purpose.
textbox_scan_step_x = 1.0 # mm
textbox_scan_step_y = 1.0 # mm
textbox_scan_width = 2.0 # mm
textbox_scan_height = 2.0 # mm

# The pixel coverage of the scan area that is required for it to
# be considered writing.
textbox_scan_coverage = 0.06

# Minimum size in mm for a textbox to be considered filled in.
# This is usefull because otherwise small dirt dots will be considered writing.
textbox_minimum_writing_width = 5 # mm
textbox_minimum_writing_height = 5 # mm

# Distance to stay away from the outline in mm so that it will not be detected
# as handwriting. The "uncorrected" value is used when the corners have not been
# detected correctly, which should hopefully never happen.
textbox_scan_padding = 0.5 # mm
textbox_scan_uncorrected_padding = 1.5 # mm

# Padding tha will always be added to the textbox size if content has been recognized.
# This is important because we want the outline to be visible if someone wrote
# over the outline(so that the reader can see that the software worked correctly).
textbox_extra_padding = 0.5 # mm


# Image ===================================================

# The width of the lines in the scanned image.
# All lines are 1pt wide(1/72 inch or 25.4/72 mm)
image_line_width = 24.4/72. # mm
# The coverage that the line needs to have for recognition
image_line_coverage = 0.35


# Allowed characters in code 128 barcodes (only ascii for now)
c128_chars = [chr(i) for i in range(32, 127)] #+ [u'È', u'É', u'Ê', u'Ë', u'Ì', u'Í', u'Î', u'Ï', u'Ð', u'Ñ', u'Ò', u'Ó']


# External commands =======================================
#: The binary used to compile latex documents.
latex_engine = "pdflatex"

#: A function that is called after fork and before exec of the latex engine.
#: This is useful when e.g. the LateX environment should be secured.
latex_preexec_hook = None



# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, 2011, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This modules contains the functionality to create reports using LaTeX.
"""

import os
import tempfile
import shutil
import glob

from sdaps import model

from sdaps import clifilter
from sdaps import template
from sdaps import matrix
from sdaps import paths
from sdaps import defs
from sdaps.utils import latex

from sdaps.utils import paper
from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

from . import buddies
import codecs


def report(survey, filter, filename=None, papersize=None, small=0, suppress=None, tex_only=False):
    assert isinstance(survey, model.survey.Survey)

    # compile clifilter
    filter = clifilter.clifilter(survey, filter)

    # First: calculate buddies

    # init buddies
    survey.questionnaire.calculate.init()

    # iterate over sheets
    survey.iterate(
        survey.questionnaire.calculate.read,
        lambda: survey.sheet.valid and filter()
    )

    # do calculations
    survey.questionnaire.calculate.calculate()

    # Temporary directory for TeX files.
    if tex_only and filename:
        tmpdir = filename

        # Create directory
        os.makedirs(tmpdir)
    else:
        tmpdir = tempfile.mkdtemp(prefix='sdaps-report-')

    # Second: report buddies

    # init buddies
    survey.questionnaire.report.init(tmpdir, small, suppress)

    # Filename of output
    if filename is None and tex_only == False:
        filename = survey.new_path('report_%i.pdf')

    try:
        # iterate over sheets
        survey.iterate(
            survey.questionnaire.report.report,
            lambda: survey.sheet.valid and filter(),
            tmpdir
        )

        # Copy class and dictionary files
        if paths.local_run:
            cls_file = os.path.join(paths.source_dir, 'tex', 'sdapsreport.cls')
            dict_files = os.path.join(paths.build_dir, 'tex', '*.dict')
            dict_files = glob.glob(dict_files)
        else:
            cls_file = os.path.join(paths.prefix, 'share', 'sdaps', 'tex', 'sdapsreport.cls')
            dict_files = os.path.join(paths.prefix, 'share', 'sdaps', 'tex', '*.dict')
            dict_files = glob.glob(dict_files)

        shutil.copyfile(cls_file, os.path.join(tmpdir, 'sdapsreport.cls'))
        for dict_file in dict_files:
            shutil.copyfile(dict_file, os.path.join(tmpdir, os.path.basename(dict_file)))

        texfile = codecs.open(os.path.join(tmpdir, 'report.tex'), 'w', 'utf-8')

        author = _('author|Unknown')

        extra_info = []
        for key, value in survey.info.items():
            if key == 'Author':
                author = value
                continue

            extra_info.append('\\addextrainfo{%(key)s}{%(value)s}' % {'key': key, 'value': value})

        extra_info = '\n'.join(extra_info)
        texfile.write(r"""\documentclass[%(language)s,%(papersize)s]{sdapsreport}

    \usepackage{ifxetex}
    \ifxetex
    \else
      \usepackage[utf8]{inputenc}
    \fi
    \usepackage[%(language)s]{babel}

    \title{%(title)s}
    \subject{%(title)s}
    \author{%(author)s}

    \addextrainfo{%(turned_in)s}{%(count)i}
    %(extra_info)s

    \begin{document}

    \maketitle

    """ % {'language': _('tex language|english'),
           'title': _('sdaps report'),
           'turned_in': _('Turned in Questionnaires'),
           'title': survey.title,
           'author': author,
           'extra_info': extra_info,
           'count': survey.questionnaire.calculate.count,
           'papersize' : paper.get_tex_papersize(papersize)})

        survey.questionnaire.report.write(texfile, tmpdir)

        texfile.write(r"""
    \end{document}
    """)
        texfile.close()

        if tex_only:
            print(_("The TeX project with the report data is located at '%s'.") % tmpdir)
            return

        print(_("Running %s now multiple times to generate the report.") % survey.defs.engine)
        latex.compile(survey.defs.engine, 'report.tex', cwd=tmpdir)

        if not os.path.exists(os.path.join(tmpdir, 'report.pdf')):
            print(_("Error running \"%s\" to compile the LaTeX file.") % survey.defs.engine)
            raise AssertionError('PDF file not generated')

        shutil.move(os.path.join(tmpdir, 'report.pdf'), filename)

    except:
        print(_("An occured during creation of the report. Temporary files left in '%s'." % tmpdir))

        raise

    shutil.rmtree(tmpdir)




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This modules contains the functionality to create PDF based reports.
"""

import os

from reportlab import platypus

from sdaps import model

from sdaps import clifilter
from sdaps import template
from sdaps import matrix
from sdaps.utils import paper

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

from . import buddies


def report(survey, filter, filename=None, papersize=None, small=0, suppress=None):
    assert isinstance(survey, model.survey.Survey)

    # compile clifilter
    filter = clifilter.clifilter(survey, filter)

    # First: calculate buddies

    # init buddies
    survey.questionnaire.calculate.init()

    # iterate over sheets
    survey.iterate(
        survey.questionnaire.calculate.read,
        lambda: survey.sheet.valid and not survey.sheet.empty and filter()
    )

    # do calculations
    survey.questionnaire.calculate.calculate()

    # Second: report buddies

    # init buddies
    survey.questionnaire.report.init(small, suppress)

    # iterate over sheets
    survey.iterate(
        survey.questionnaire.report.report,
        lambda: survey.sheet.valid and filter()
    )

    # create story
    story = template.story_title(
        survey,
        {
            _('Turned in Questionnaires'): survey.questionnaire.calculate.count,
        }
    )
    story.extend(survey.questionnaire.report.story())

    # create report(out of story)
    if filename is None:
        filename = survey.new_path('report_%i.pdf')
    subject = []
    for key, value in survey.info.items():
        subject.append('%(key)s: %(value)s' % {'key': key, 'value': value})
    subject = '\n'.join(subject)

    papersize = paper.get_reportlab_papersize(papersize)

    doc = template.DocTemplate(
        filename,
        _('sdaps report'),
        {
            'title': survey.title,
            'subject': subject,
        },
        papersize=papersize
    )
    doc.build(story)


def stats(survey, filter, filename=None, small=0, suppress=None):
    if filename is None:
        filename = survey.new_path('report_%i.pdf')

    # do a report
    report(survey, filter, filename=filename, small=small)

    # save reference (to highlight large differences)
    survey.questionnaire.calculate.reference()

    # do a report for every filter
    for i, subfilter in enumerate(survey.questionnaire.report.filters()):
        if filter is None:
            filt = subfilter
        else:
            filt = "(%s) and (%s)" % (filter, subfilter)
        report(
            survey, filt,
            filename='%s_%i %s.pdf' % (filename.split('.')[0], i, subfilter),
            small=small,
            suppress=suppress)




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
This module defines some decorators, helping to implement sdaps-scripts.
To register a function as a sdaps-script(callable from the command line), use
@register
'''

import sys
import os
import functools
import argparse

from . import __version__
from . import log
from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

# Create parser

if "sphinx" in sys.argv[0]:
    prog = "sdaps"
else:
    prog = None

description = _("SDAPS -- Paper based survey tool.")
epilog = None
parser = argparse.ArgumentParser(description=description, epilog=epilog, prog=prog)

parser.add_argument('--version',
    help=_('Display version and exit'),
    action='version',
    version="%%(prog)s %s" % __version__)

# Set required as an attribute rather than kwarg so that it works with python <3.7
subparsers = parser.add_subparsers(help=_("command list|Commands:"), dest='command')
subparsers.required = True

def add_subparser(*args, **kwargs):
    parser = subparsers.add_parser(*args, **kwargs)

    return parser

def add_project_argument(parser):
    parser.add_argument('project', type=str, help=_("project directory|The SDAPS project."))

def add_project_subparser(*args, **kwargs):
    parser = add_subparser(*args, **kwargs)

    add_project_argument(parser)

    return parser

def doc(docstring):
    '''decorator to add a docstring to a function.

    When using normal Python docstring syntax it cannot be generated
    dynamically. Using this one can for example add translations.

    >>> @doc(_(u'docstring'))
    >>> def function(*args, **kwargs):
    >>>    pass
    '''

    def decorator(function):
        function.__doc__ = docstring
        return function
    return decorator

def connect(parser, name=None):
    '''decorator to connect an already prepared parser to call into a function.

    This function initilizes the _func and _name properties for the parser to
    the given function, and its name. It also sets the functions docstring to
    be the parsers help. This way the help string appears in the sphinx
    documentation for the function.

    >>> @script.connect(parser)
    >>> def add(cmdline):
    >>>     pass
    '''

    def decorator(function):
        # Use the function name as a fallback, it should be the same usually.
        if name is None:
            local_name = function.__name__
        else:
            local_name = name

        parser.set_defaults(_func=function, _name=local_name)

        function.__doc__ = parser.format_help()

        return function

    return decorator

def logfile(function):
    '''open the logfile when running the function and close it afterwards.

    >>> @logfile
    >>> def function(survey_dir, *args, **kwargs):
    >>>     pass

    @logfile will open survey_dir/log as a logfile when function is called and
    close it, when function finishes.
    '''
    def decorated_function(cmdline):
        log.logfile.open(os.path.join(cmdline['project'], 'log'))
        result = function(cmdline)
        log.logfile.close()
        return result

    functools.update_wrapper(decorated_function, function)

    return decorated_function





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2012, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from sdaps import model
from sdaps import script

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext


parser = script.add_project_subparser("annotate",
    help=_("Annotate the questionnaire and show the recognized positions."),
    description=_("""This command is mainly a debug utility. It creates an
    annotated version of the questionnaire, with the information that SDAPS
    knows about it overlayed on top."""))

@script.connect(parser)
@script.logfile
def annotate(cmdline):
    from sdaps.annotate import annotate

    survey = model.survey.Survey.load(cmdline['project'])

    return annotate(survey)




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gettext


def ugettext(string):
    '''gettext for unicode objects
    '''
    translation = gettext.gettext(string.encode('UTF-8')).decode('UTF-8')

    return translation.split('|', 1)[-1]


def ungettext(singular, plural, n):
    '''ngettext for unicode objects
    '''
    return gettext.ngettext(
        singular.encode('UTF-8'),
        plural.encode('UTF-8'),
        n).decode('UTF-8')



# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sdaps import model
from sdaps import script

import os
import sys

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

parser = script.add_project_subparser("ids",
    help=_("Export and import questionnaire IDs."),
    description=_("""This command can be used to import and export questionnaire
    IDs. It only makes sense in projects where such an ID is printed on the
    questionnaire. Note that you can also add IDs by using the stamp command,
    which will give you the PDF at the same time."""))
parser.add_argument('-o', '--output',
    help=_("Filename to store the data to (default: ids_%%i)"))
parser.add_argument('-a', '--add',
    metavar="FILE",
    help=_("Add IDs to the internal list from the specified file."))

@script.connect(parser)
@script.logfile
def ids(cmdline):
    survey = model.survey.Survey.load(cmdline['project'])

    if cmdline['add']:
        if cmdline['add'] == '-':
            ids = sys.stdin
        else:
            ids = open(cmdline['add'], 'r')

        to_add = []
        for line in ids.readlines():
            # Skip empty lines
            if line == "":
                continue

            line = line
            line = line.strip('\r\n')
            to_add.append(survey.validate_questionnaire_id(line))

        survey.questionnaire_ids += to_add
        survey.save()

    else:
        if cmdline['output']:
            if cmdline['output'] == '-':
                outfd = os.dup(sys.stdout.fileno())
                ids = os.fdopen(outfd, 'w')
            else:
                filename = cmdline['output']
                ids = open(filename, 'w')
        else:
            filename = survey.new_path('ids_%i')
            ids = open(filename, 'w')

        for id in survey.questionnaire_ids:
            ids.write(str(id))
            ids.write('\n')

        if ids != sys.stdout:
            ids.close()





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sdaps import model
from sdaps import script

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext

parser = script.add_project_subparser("stamp",
    help=_("Add marks for automatic processing."),
    description=_("""This command creates the printable document. Depending on
    the projects setting you are required to specifiy a source for questionnaire
    IDs."""))

parser.add_argument('-r', '--random',
    metavar="N",
    help=_("If using questionnaire IDs, create N questionnaires with randomized IDs."),
    type=int)
parser.add_argument('-f', '--file',
    help=_("If using questionnaire IDs, create questionnaires from the IDs read from the specified file."))
parser.add_argument('--existing',
    action="store_true",
    help=_("If using questionnaire IDs, create questionnaires for all stored IDs."))

parser.add_argument('-o', '--output',
    help=_("Filename to store the data to (default: stamp_%%i.pdf)"))

@script.connect(parser)
@script.logfile
def stamp(cmdline):
    from sdaps import stamp

    survey = model.survey.Survey.load(cmdline['project'])

    if cmdline['output'] is None:
        output = survey.new_path('stamped_%i.pdf')
    else:
        output = cmdline['output']

    return stamp.stamp(survey, output, cmdline)





# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This modules contains a helper function to allow writing filter expressions
on the command line of sdaps. Using this it is for example possible to create
a report that only contains a subset of all filled out sheets.
"""


class Locals(object):

    def __init__(self, survey):
        self.survey = survey
        self.qobjects = dict([
            (qobject.id_filter(), qobject)
            for qobject in self.survey.questionnaire.qobjects
        ])

    def __getitem__(self, key):
        if key in self.qobjects:
            return self.qobjects[key].get_answer()
        elif key in ['survey_id', 'questionnaire_id', 'global_id', 'valid', 'quality', 'recognized', 'verified', 'complete']:
            return getattr(self.survey.sheet, key)
        else:
            raise KeyError


def clifilter(survey, expression):
    if expression is None or expression.strip() == '':
        return lambda: True

    exp = compile(expression, '<string>', 'eval')
    globals = __builtins__
    locals = Locals(survey)
    return lambda: eval(exp, globals, locals)




# -*- coding: utf-8 -*-
# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2008, Christoph Simon <post@christoph-simon.eu>
# Copyright(C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sdaps import template

from sdaps.utils.ugettext import ugettext, ungettext
_ = ugettext


def cover(survey, output=None):

    story = template.story_title(survey)
    subject = []
    for key, value in survey.info.items():
        subject.append('%(key)s: %(value)s' % {'key': key, 'value': value})
    subject = '\n'.join(subject)

    if output:
        filename = output
    else:
        filename = survey.new_path('cover_%i.pdf')

    doc = template.DocTemplate(
        filename,
        _('SDAPS questionnaire'),
        {
            'title': survey.title,
            'subject': subject
        }
    )
    doc.build(story)





# SDAPS - Scripts for data acquisition with paper based surveys
# Copyright(C) 2014, Michael Batchelor <batchelor@nationbuilder.com>
# Copyright(C) 2014, Chuck Collins <chuck@nationbuilder.com>
# Copyright(C) 2014, Jacob Green <jacob@nationbuilder.com>
# Copyright(C) 2014, Dustin Ngo <dustin@nationbuilder.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sdaps import defs

from sdaps import model
from sdaps.utils.exceptions import RecognitionError
from sdaps.utils.barcode import read_barcode


class Image(model.buddy.Buddy, metaclass=model.buddy.Register):

    name = 'style'
    obj_class = model.sheet.Image

    def get_page_rotation(self):
        code = self.find_bottom_right_barcode()
        if code is None:
            # Well, that failed, so try to search the upper left corner instead
            code = self.find_top_left_barcode()

            if code is not None:
                return True
            else:
                return None
        else:
            return False

    def get_page_number(self):
        code = self.find_bottom_right_barcode()

        if code is None or (not code.isdigit() and len(code) < 4):
            return None

        return int(code[-4:])

    def get_survey_id(self):
        code = self.find_bottom_right_barcode()

        if code is None or not code.isdigit() or len(code) <= 4:
            return None

        return int(code[:-4])

    def get_questionnaire_id(self):
        return self.find_bottom_left_barcode()

    def get_global_id(self):
        return self.find_bottom_center_barcode()

    def find_bottom_right_barcode(self):
      return read_barcode(self.obj.surface.surface, self.obj.matrix.mm_to_px(),
                 self.paper_width() * 0.75,
                 self.paper_height() * 0.75,
                 self.paper_width() * 0.25,
                 self.paper_height() * 0.25,
                 "QRCODE")

    def find_top_left_barcode(self):
      return read_barcode(self.obj.surface.surface, self.obj.matrix.mm_to_px(),
                     0, 0,
                     self.paper_width() * 0.25,
                     self.paper_height() * 0.25,
                     "QRCODE")

    def find_bottom_left_barcode(self):
      return read_barcode(self.obj.surface.surface, self.obj.matrix.mm_to_px(),
                     0,
                     self.paper_height() * 0.75,
                     self.paper_width() * 0.25,
                     self.paper_height() * 0.25,
                     "QRCODE")

    def find_bottom_center_barcode(self):
      return read_barcode(self.obj.surface.surface, self.obj.matrix.mm_to_px(),
                     self.paper_width() * 0.375,
                     self.paper_height() * 0.75,
                     self.paper_width() * 0.25,
                     self.paper_height() * 0.25,
                     "QRCODE")

    def paper_width(self):
      return self.obj.sheet.survey.defs.paper_width

    def paper_height(self):
      return self.obj.sheet.survey.defs.paper_height
